from mitreattack.navlayers.core.exceptions import BadInput, handler, typeChecker, loadChecker, UNSETVALUE, \
    UnknownTechniqueProperty, BadType, MissingParameters
from mitreattack.navlayers.core.metadata import Metadata, MetaDiv
from mitreattack.navlayers.core.objlink import Link, LinkDiv
from mitreattack.navlayers.core.helpers import handle_object_placement


class Technique:
    def __init__(self, tID):
        """
            Initialization - Creates a technique object

            :param tID: The techniqueID associated with this technique object
        """
        self.techniqueID = tID
        self.__tactic = UNSETVALUE
        self.__comment = UNSETVALUE
        self.__enabled = UNSETVALUE
        self.__score = UNSETVALUE
        self.__color = UNSETVALUE
        self.__metadata = UNSETVALUE
        self.__showSubtechniques = UNSETVALUE
        self.__aggregateScore = UNSETVALUE
        self.__links = UNSETVALUE

    @property
    def techniqueID(self):
        return self.__techniqueID

    @techniqueID.setter
    def techniqueID(self, techniqueID):
        typeChecker(type(self).__name__, techniqueID, str, "techniqueID")
        if not techniqueID.startswith('T'):
            handler(type(self).__name__, '{} not a valid value for techniqueID'
                    .format(techniqueID))
            raise BadInput
        else:
            self.__techniqueID = techniqueID

    @property
    def tactic(self):
        if self.__tactic != UNSETVALUE:
            return self.__tactic

    @tactic.setter
    def tactic(self, tactic):
        typeChecker(type(self).__name__, tactic, str, "tactic")
        self.__tactic = tactic

    @property
    def comment(self):
        if self.__comment != UNSETVALUE:
            return self.__comment

    @comment.setter
    def comment(self, comment):
        typeChecker(type(self).__name__, comment, str, "comment")
        self.__comment = comment

    @property
    def enabled(self):
        if self.__enabled != UNSETVALUE:
            return self.__enabled

    @enabled.setter
    def enabled(self, enabled):
        typeChecker(type(self).__name__, enabled, bool, "enabled")
        self.__enabled = enabled

    @property
    def score(self):
        if self.__score != UNSETVALUE:
            return self.__score

    @score.setter
    def score(self, score):
        try:
            typeChecker(type(self).__name__, score, int, "score")
            self.__score = score
        except BadType:
            typeChecker(type(self).__name__, score, float, "score")
            self.__score = int(score)

    @property
    def color(self):
        if self.__color != UNSETVALUE:
            return self.__color

    @color.setter
    def color(self, color):
        typeChecker(type(self).__name__, color, str, "color")
        self.__color = color

    @property
    def metadata(self):
        if self.__metadata != UNSETVALUE:
            return self.__metadata

    @metadata.setter
    def metadata(self, metadata):
        typeChecker(type(self).__name__, metadata, list, "metadata")
        self.__metadata = []

        try:
            for entry in metadata:
                try:
                    if isinstance(entry, Metadata) or isinstance(entry, MetaDiv):
                        loadChecker(type(self).__name__, entry.get_dict(), ['name', 'value'], "metadata")
                        self.__metadata.append(entry)
                    elif isinstance(entry, dict):
                        loadChecker(type(self).__name__, entry, ['name', 'value'], "metadata")
                        if entry['name'] == "DIVIDER":
                            self.__metadata.append(MetaDiv(active=entry['value']))
                        else:
                            self.__metadata.append(Metadata(name=entry['name'], value=entry['value']))
                    else:
                        pass  # Object in the list was not of Metadata or MetaDiv classes
                except MissingParameters as e:
                    handler(
                        type(self).__name__,
                        'Metadata {} is missing parameters: {}. Skipping.'.format(entry, e)
                    )
        except KeyError as e:
            handler(type(self).__name__, 'Metadata {} is missing parameters: '
                                         '{}. Unable to load.'
                    .format(entry, e))

    @property
    def showSubtechniques(self):
        if self.__showSubtechniques != UNSETVALUE:
            return self.__showSubtechniques

    @showSubtechniques.setter
    def showSubtechniques(self, showSubtechniques):
        typeChecker(type(self).__name__, showSubtechniques, bool,
                    "showSubtechniques")
        self.__showSubtechniques = showSubtechniques

    @property
    def aggregateScore(self):
        if self.__aggregateScore != UNSETVALUE:
            return self.__aggregateScore

    @aggregateScore.setter
    def aggregateScore(self, aggregateScore):
        typeChecker(type(self).__name__, aggregateScore, int, "aggregate")
        self.__aggregateScore = aggregateScore

    @property
    def links(self):
        if self.__links != UNSETVALUE:
            return self.__links

    @links.setter
    def links(self, links):
        typeChecker(type(self).__name__, links, list, "links")
        if not handle_object_placement(self.__links, links, Link):
            self.__links = []
        entry = ""
        try:
            for entry in links:
                if isinstance(entry, Link):
                    loadChecker(type(self).__name__, entry.get_dict(), ['label', 'url'], "link")
                    self.__links.append(entry)
                elif isinstance(entry, LinkDiv):
                    loadChecker(type(self).__name__, entry.get_dict(), ['name', 'value'], "linkdiv")
                    self.__links.append(entry)
                elif isinstance(entry, dict):
                    if 'name' in entry and entry['name'] == "DIVIDER":
                        loadChecker(type(self).__name__, entry, ['name', 'value'], "linkdiv")
                        self.__links.append(LinkDiv(active=entry['value']))
                    else:
                        loadChecker(type(self).__name__, entry, ['label', 'url'], "link")
                        self.__links.append(Link(label=entry['label'], url=entry['url']))
                else:
                    pass
        except KeyError as e:
            handler(type(self).__name__, 'Link {} is missing parameters: '
                                         '{}. Unable to load.'
                    .format(entry, e))

    def _loader(self, data):
        """
            INTERNAL: Acts a middleman for loading values into the technique
                object from a dict representation
            :param data: A dict describing the technique
            :raises UnknownTechniqueProperty: An error indicating that an
                unexpected property was found on the technique
        """
        for entry in data.keys():
            if entry == 'techniqueID':
                pass
            elif entry == 'tactic':
                self.tactic = data[entry]
            elif entry == 'comment':
                self.comment = data[entry]
            elif entry == 'enabled':
                self.enabled = data[entry]
            elif entry == 'score':
                self.score = data[entry]
            elif entry == 'color':
                self.color = data[entry]
            elif entry == 'metadata':
                self.metadata = data[entry]
            elif entry == 'showSubtechniques':
                self.showSubtechniques = data[entry]
            elif entry == 'links':
                self.links = data[entry]
            elif entry == 'aggregateScore':
                self.aggregateScore = data[entry]
            else:
                handler(type(self).__name__, "Unknown technique property: {}"
                        .format(entry))
                raise UnknownTechniqueProperty

    def get_dict(self):
        """
            Converts the currently loaded data into a dict
            :returns: A dict representation of the local technique object
        """
        dset = vars(self)
        temp = {}
        for key in dset:
            entry = key.split(type(self).__name__ + '__')[-1]
            if dset[key] != UNSETVALUE:
                if entry != 'metadata' and entry != 'links':
                    temp[entry] = dset[key]
                else:
                    temp[entry] = [x.get_dict() for x in dset[key]]
        return temp



import argparse
import json
from tabulate import tabulate
from dateutil.parser import isoparse


class IndexToMarkdown:
    @staticmethod
    def index_to_markdown(indexobj):
        """
        convert the index object [indexobj] (dict) to a markdown string and return the markdown string
        """
        lines = ["<!-- begin index generated by mitreattack-python.navlayers.collections.col_to_md.py -->",
                 f"## {indexobj['name']}", f"\n{indexobj['description']}\n", f"### {indexobj['name']} Collections\n"]

        for collection in indexobj['collections']:
            lines.append(f"#### {collection['name']}")
            lines.append(f"\n{collection['description']}\n")
            format_versions = list(map(lambda v: [
                f"[{collection['name']} v{v['version']}]({v['url']})",
                isoparse(v['modified']).strftime("%d %B %Y")
            ], collection["versions"]))

            lines.append(tabulate(format_versions, headers=['version', 'released'], tablefmt="github"))
            lines.append("")

        lines.append("<!-- end index generated by  mitreattack-python.navlayers.collections.col_to_md.py -->")

        return "\n".join(lines)


def main(args):
    with open(args.index, "r", encoding="utf-16") as f:
        index = json.load(f)
        with open(args.output, "w", encoding="utf-16") as f2:
            f2.write(IndexToMarkdown.index_to_markdown(index))


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Print a markdown string to std-out representing a collection index"
    )
    parser.add_argument("-i",
                        "--index",
                        type=str,
                        default="index.json",
                        help="the collection index file to convert to markdown"
                        )
    parser.add_argument("-o",
                        "--output",
                        type=str,
                        default="index.md",
                        help="markdown output file"
                        )
    argv = parser.parse_args()
    main(argv)



from stix2 import Filter

from mitreattack.navlayers.exporters.matrix_gen import MatrixGen
from mitreattack.navlayers.core.exceptions import BadInput, typeChecker, categoryChecker
from mitreattack.navlayers.core.layer import Layer
from mitreattack.navlayers.generators.gen_helpers import remove_revoked_depreciated, construct_relationship_mapping, \
     MITRE_ATTACK_DOMAIN_STRINGS, build_data_strings


class UnableToFindTechnique(Exception):
    pass


class OverviewLayerGenerator:
    """Generates a Layer file that provides an overview of entities related to each technique"""
    def __init__(self, source, domain='enterprise', resource=None):
        """
        Initialize the Generator
        :param source: Which source to use for data (local, taxii [server], or [remote] ATT&CK Workbench)
        :param domain: Which domain to use during generation
        :param resource: string path to local cache of stix data (local) or url of Workbench to connect to (remote)
        """
        self.matrix_handle = MatrixGen(source, resource)
        self.domain = domain
        try:
            self.source_handle = self.matrix_handle.collections[domain]
        except KeyError:
            print(f"[UsageGenerator] - unable to load collection {domain} (current source = {source}).")
            raise BadInput
        tl = remove_revoked_depreciated(self.source_handle.query([Filter('type', '=', 'attack-pattern')]))
        self.mitigation_objects = self.source_handle.query([Filter('type', '=', 'course-of-action')])
        complete_relationships = self.source_handle.query([Filter('type', '=', 'relationship'),
                                                           Filter('relationship_type', '=', 'uses')])
        complete_relationships.extend(self.source_handle.query([Filter('type', '=', 'relationship'),
                                                                Filter('relationship_type', '=', 'mitigates')]))
        complete_relationships.extend(self.source_handle.query([Filter('type', '=', 'relationship'),
                                                                Filter('relationship_type', '=', 'detects')]))

        self.sources = self.source_handle.query([Filter('type', '=', 'x-mitre-data-source')])
        self.components = self.source_handle.query([Filter('type', '=', 'x-mitre-data-component')])
        self.source_mapping = build_data_strings(self.sources, self.components)
        # Contains relationship mapping [stix id] -> [relationships associated with that stix id for each type]
        self.simplifier = {"course-of-action": dict(), "tool": dict(),
                           "malware": dict(), "intrusion-set": dict(),
                           "x-mitre-data-component": dict()}

        # Scan through all relationships to identify ones that target attack techniques (attack-pattern). Then, sort
        # these into the mapping dictionary by what kind of stix object they are (tool, course-of-action, etc.)
        for entry in complete_relationships:
            if entry['target_ref'].startswith('attack-pattern--'):
                construct_relationship_mapping(self.simplifier[entry['source_ref'].split('--')[0]], entry)
        self.simplifier['software'] = self.simplifier['malware']
        self.simplifier['software'].update(self.simplifier['tool'])  # get combination of malware/tool
        self.simplifier['datasource'] = self.simplifier['x-mitre-data-component']

        self.tech_listing = dict()
        self.tech_no_tactic_listing = dict()
        for entry in tl:
            tacs = [None]
            xid = None
            for ref in entry.external_references:
                if ref.source_name in MITRE_ATTACK_DOMAIN_STRINGS:
                    xid = ref.external_id
                    break
            for phase in entry.kill_chain_phases:
                if phase.kill_chain_name in MITRE_ATTACK_DOMAIN_STRINGS:
                    tacs.append(phase.phase_name)
            for xphase in tacs:
                self.tech_listing[(xid, xphase)] = entry
            self.tech_no_tactic_listing[xid] = entry

    def get_groups(self, relationships):
        """
        Sort Groups out of relationships
        :param relationships: List of all related relationships to a given technique
        :return: length of matched groups, list of group names
        """
        list_of_groups = []
        for relationship in relationships:
            if relationship.source_ref.startswith('intrusion-set--'):
                list_of_groups.append(relationship)
        group_objects = self.source_handle.query([Filter('type', '=', 'intrusion-set'),
                                                  Filter('id', 'in', [r.source_ref for r in list_of_groups])])
        names = [x.name for x in group_objects]
        return len(names), names

    def get_software(self, relationships):
        """
        Sort software out of relationships
        :param relationships: List of all related relationships to a given technique
        :return: length of matched software, list of software names
        """
        list_of_softwares = []
        for relationship in relationships:
            if relationship.source_ref.startswith('malware--') or relationship.source_ref.startswith('tool--'):
                list_of_softwares.append(relationship.source_ref)
        software_listing = self.source_handle.query([Filter('type', '=', 'malware'), Filter('type', '=', 'tool')])
        software_objects = []
        for soft in software_listing:
            if soft.id in list_of_softwares:
                software_objects.append(soft)
        names = [x.name for x in software_objects]
        return len(names), names

    def get_mitigations(self, relationships):
        """
        Sort mitigations out of relationships
        :param relationships: List of all related relationships to a given technique
        :return: length of matched mitigations, list of mitigation names
        """
        names = [x.name for x in self.mitigation_objects if x.id in relationships]
        return len(names), names

    def get_datasources(self, relationships):
        """
        Sort datasources/datacomponents out of relationships
        :param relationships: List of all related relationships to a given technique
        :return: length of matched datasources/datacomponents, list of associated names
        """
        names = [self.source_mapping[x.source_ref] for x in relationships]
        return len(names), names

    def get_matrix_template(self):
        """
        Build the raw dictionary form matrix layer object
        :return: dictionary representing all entries in the matrix layer
        """
        construct = list()
        full_matrix_listing = self.matrix_handle.get_matrix(self.domain)
        for tactic in full_matrix_listing:
            for tech in tactic.techniques:
                construct.append(dict(techniqueID=tech.id, score=0,
                                      tactic=self.matrix_handle.convert(tactic.tactic.name)))
            for tech_key in tactic.subtechniques:
                for subtech in tactic.subtechniques[tech_key]:
                    construct.append(dict(techniqueID=subtech.id, score=0,
                                          tactic=self.matrix_handle.convert(tactic.tactic.name)))
        return construct

    def get_technique_obj(self, techniqueID, tactic):
        """
        Extract the matching technique object from the tech_listing
        :param techniqueID: the technique object's id
        :param tactic: optional tactic for the technique object (shortname - ex. "reconnaissance")
        :return: the matching technique object (or a UnableToFindTechnique exception)
        """
        try:
            return self.tech_listing[(techniqueID, tactic)]
        except KeyError:
            pass  # didn't find a specific match for that combo, let's drop the tactic and see what we get
        try:
            return self.tech_no_tactic_listing[techniqueID]
        except KeyError:
            raise UnableToFindTechnique

    def update_template(self, obj_type, complete_tech_listing):
        """
        Update an existing dictionary of layer techniques with the appropriate matching objects
        :param obj_type: the type of object to update the data with
        :param complete_tech_listing: 'clean' technique dictionary template
        :return: Updated technique dictionary
        """
        temp = complete_tech_listing
        for entry in temp:
            tech = self.get_technique_obj(entry['techniqueID'], entry['tactic'])
            score = 0
            listing = []
            if obj_type == 'group':
                try:
                    related = self.simplifier['intrusion-set'][tech.id]
                    score, listing = self.get_groups(related)
                except KeyError:
                    pass
            elif obj_type == 'software':
                try:
                    related = self.simplifier['software'][tech.id]
                    score, listing = self.get_software(related)
                except KeyError:
                    pass
            elif obj_type == "mitigation":
                try:
                    related = self.simplifier['course-of-action'][tech.id]
                    score, listing = self.get_mitigations(related)
                except KeyError:
                    pass  # we don't have any matches for this one
            elif obj_type == "datasource":
                try:
                    related = self.simplifier['datasource'][tech.id]
                    score, listing = self.get_datasources(related)
                except KeyError:
                    pass
            entry['score'] = score
            entry['comment'] = ', '.join(listing)
        return temp

    def generate_layer(self, obj_type):
        """
        Generate a layer
        :param obj_type: the type of object data to compute over (group, software, or mitigation)
        :return: layer object with annotated techniques
        """
        typeChecker(type(self).__name__, obj_type, str, "type")
        categoryChecker(type(self).__name__, obj_type, ["group", "software", "mitigation", "datasource"], "type")
        initial_list = self.get_matrix_template()
        updated_list = self.update_template(obj_type, initial_list)
        if obj_type == "group":
            p_name = "groups"
            r_type = "using"
        elif obj_type == "software":
            p_name = "software"
            r_type = "using"
        elif obj_type == "datasource":
            p_name = "data sources"
            r_type = "detecting"
        else:  # mitigation case
            p_name = "mitigations"
            r_type = "mitigating"
        desc = f"Overview of techniques used by {p_name}. Score is the number of {p_name} " \
               f"{r_type} the technique, and comment lists the {r_type} {p_name}"
        raw_layer = dict(name=f"{p_name} overview", domain='enterprise-attack', description=desc)
        raw_layer['techniques'] = updated_list
        output_layer = Layer(raw_layer)
        return output_layer



import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    import drawSvg as draw

from mitreattack.navlayers.exporters.matrix_gen import MatrixGen
from mitreattack.navlayers.exporters.svg_objects import G, SVG_HeaderBlock, SVG_Technique, Text, convertToPx, \
     _optimalFontSize
from mitreattack.navlayers.core.gradient import Gradient
from mitreattack.navlayers.core.filter import Filter


class BadTemplateException(Exception):
    pass


class SvgTemplates:
    def __init__(self, source='taxii', domain='enterprise', resource=None):
        """
            Initialization - Creates a SvgTemplate object

            :param domain: Which domain to utilize for the underlying matrix layout
            :param source: Use the taxii server or local data
            :param resource: string path to local cache of stix data (local) or url of workbench to reach out
                                to (remote)
        """
        muse = domain
        if muse.startswith('mitre-'):
            muse = domain[6:]
        if muse.endswith('-attack'):
            muse = domain[:-7]
        if muse in ['enterprise', 'mobile']:
            self.mode = muse
            self.h = MatrixGen(source=source, resource=resource)
            self.lhandle = None
        else:
            raise BadTemplateException

    def _build_headers(self, name, config, domain='Enterprise', version='8', desc=None, filters=None, gradient=None):
        """
            Internal - build the header blocks for the svg

            :param name: The name of the layer being exported
            :param config: SVG Config object
            :param domain: The layer's domain
            :param version: The layer's version
            :param desc: Description of the layer being exported
            :param filters: Any filters applied to the layer being exported
            :param gradient: Gradient information included with the layer
            :return: Instantiated SVG header
        """
        max_x = convertToPx(config.width, config.unit)
        max_y = convertToPx(config.height, config.unit)
        header_height = convertToPx(config.headerHeight, config.unit)
        ff = config.font
        d = draw.Drawing(max_x, max_y, origin=(0, -max_y), displayInline=False)
        psych = 0
        overlay = None
        if config.showHeader:
            border = convertToPx(config.border, config.unit)
            root = G(tx=border, ty=border, style='font-family: {}'.format(ff))

            header = G()
            root.append(header)
            b1 = G()
            header.append(b1)

            header_count = 0
            showAgg = False
            if config.showAbout:
                header_count += 1
            if config.showFilters:
                header_count += 1
            if config.showDomain:
                header_count += 1
            if config.showLegend and gradient is not False and config.legendDocked:
                header_count += 1
            if self.lhandle.layout:
                if self.lhandle.layout.showAggregateScores:
                    showAgg = True
                    header_count += 1
                if config.showFilters:
                    header_count -= 1

            operation_x = (max_x - border) - (1.5 * border * (header_count - 1)) - border
            if header_count > 0:
                header_width = operation_x / header_count
                if config.showAbout:
                    if desc is not None:
                        g = SVG_HeaderBlock().build(height=header_height, width=header_width, label='about',
                                                    t1text=name, t2text=desc, config=config)
                    else:
                        g = SVG_HeaderBlock().build(height=header_height, width=header_width, label='about',
                                                    t1text=name, config=config)
                    b1.append(g)
                    psych += 1
                if config.showDomain:
                    if domain.startswith('mitre-'):
                        domain = domain[6:].capitalize()
                    if domain.endswith('-attack'):
                        domain = domain[:-7].capitalize()
                    tag = domain + ' ATT&CK v' + version
                    if config.showFilters and showAgg:
                        fi = filters
                        if fi is None:
                            fi = Filter()
                            fi.platforms = ["Windows", "Linux", "macOS"]
                        gD = SVG_HeaderBlock().build(height=header_height, width=header_width,
                                                     label='domain & platforms', t1text=tag,
                                                     t2text=', '.join(fi.platforms), config=config)
                    else:
                        gD = SVG_HeaderBlock().build(height=header_height, width=header_width, label='domain',
                                                     t1text=tag, config=config)
                    bD = G(tx=operation_x / header_count * psych + 1.5 * border * psych)
                    header.append(bD)
                    bD.append(gD)
                    psych += 1
                if config.showFilters and not showAgg:
                    fi = filters
                    if fi is None:
                        fi = Filter()
                        fi.platforms = ["Windows", "Linux",	"macOS"]
                    g2 = SVG_HeaderBlock().build(height=header_height, width=header_width, label='filters',
                                                 t1text=', '.join(fi.platforms), config=config)
                    b2 = G(tx=operation_x / header_count * psych + 1.5 * border * psych)
                    header.append(b2)
                    b2.append(g2)
                    psych += 1
                if showAgg:
                    t1 = f"showing aggregate scores using the {self.lhandle.layout.aggregateFunction} " \
                         f"aggregate function"
                    stub = "does not include"
                    if self.lhandle.layout.countUnscored:
                        stub = "includes"
                    t2 = f"{stub} unscored techniques as having a score of 0"
                    gA = SVG_HeaderBlock().build(height=header_height, width=header_width, label='aggregate',
                                                 t1text=t1, t2text=t2, config=config)
                    bA = G(tx=operation_x / header_count * psych + 1.5 * border * psych)
                    header.append(bA)
                    bA.append(gA)
                    psych += 1
                if config.showLegend and gradient is not False:
                    gr = gradient
                    if gr is None:
                        gr = Gradient(colors=["#ff6666", "#ffe766", "#8ec843"], minValue=1, maxValue=100)
                    colors = []
                    div = round((gr.maxValue - gr.minValue) / (len(gr.colors) * 2 - 1))
                    for i in range(0, len(gr.colors) * 2 - 1):
                        colors.append(
                            (gr.compute_color(int(gr.minValue + div * i)), gr.minValue + div * i))
                    colors.append((gr.compute_color(gr.maxValue), gr.maxValue))
                    if config.legendDocked:
                        b3 = G(tx=operation_x / header_count * psych + 1.5 * border * psych)
                        g3 = SVG_HeaderBlock().build(height=header_height, width=header_width, label='legend',
                                                     variant='graphic', colors=colors, config=config)
                        header.append(b3)
                        b3.append(g3)
                        psych += 1
                    else:
                        adjusted_height = convertToPx(config.legendHeight, config.unit)
                        adjusted_width = convertToPx(config.legendWidth, config.unit)
                        g3 = SVG_HeaderBlock().build(height=adjusted_height, width=adjusted_width, label='legend',
                                                     variant='graphic', colors=colors, config=config)
                        lx = convertToPx(config.legendX, config.unit)
                        if not lx:
                            lx = max_x - adjusted_width - convertToPx(config.border, config.unit)
                        ly = convertToPx(config.legendY, config.unit)
                        if not ly:
                            ly = max_y - adjusted_height - convertToPx(config.border, config.unit)
                        overlay = G(tx=lx, ty=ly)
                        if (ly + adjusted_height) > max_y or (lx + adjusted_width) > max_x:
                            print("[WARNING] - Floating legend will render partly out of view...")
                        overlay.append(g3)
            d.append(root)
        return d, psych, overlay

    def get_tactic(self, tactic, height, width, config, colors=[], scores=[], subtechs=[], exclude=[],
                   mode=(True, False)):
        """
            Build a 'tactic column' svg object

            :param tactic: The corresponding tactic for this column
            :param height: A technique block's allocated height
            :param width: A technique blocks' allocated width
            :param config: A SVG Config object
            :param colors: Default color data in case of no score
            :param scores: Score values for the dataset
            :param subtechs: List of visible subtechniques
            :param exclude: List of excluded techniques
            :param mode: Tuple describing text for techniques (Show Name, Show ID)
            :return: Instantiated tactic column (or none if no techniques were found)
        """
        offset = 0
        column = G(ty=2)
        for a in tactic.subtechniques:
            self._copy_scores(tactic.subtechniques[a], scores, tactic.tactic.name, exclude)
        for x in tactic.techniques:
            if any(x.id == y[0] and (y[1] == self.h.convert(tactic.tactic.name) or not y[1]) for y in exclude):
                continue
            self._copy_scores([x], scores, tactic.tactic.name, exclude)
            if any(x.id == y[0] and (y[1] == self.h.convert(tactic.tactic.name) or not y[1]) for y in subtechs):
                a, offset = self.get_tech(offset, mode, x, tactic=self.h.convert(tactic.tactic.name),
                                          subtechniques=tactic.subtechniques.get(x.id, []), colors=colors,
                                          config=config, height=height, width=width,
                                          subscores=tactic.subtechniques.get(x.id, []))
            else:
                a, offset = self.get_tech(offset, mode, x, tactic=self.h.convert(tactic.tactic.name),
                                          subtechniques=[], colors=colors, config=config, height=height, width=width,
                                          subscores=tactic.subtechniques.get(x.id, []))
            column.append(a)
        if len(column.children) == 0:
            return None
        return column

    def get_tech(self, offset, mode, technique, tactic, config, height, width, subtechniques=[], colors=[],
                 subscores=[]):
        """
            Retrieve a svg object for a single technique

            :param offset: The offset in the column based on previous work
            :param mode: Tuple describing display format (Show Name, Show ID)
            :param technique: The technique to build a block for
            :param tactic: The corresponding tactic
            :param config: An SVG Config object
            :param height: The allocated height of a technique block
            :param width: The allocated width of a technique block
            :param subtechniques: A list of all visible subtechniques, some of which may apply to this one
            :param colors: A list of all color overrides in the event of no score, which may apply
            :param subscores: List of all subtechniques for the (visible or not) [includes scores]
            :return: Tuple (SVG block, new offset)
        """
        # Handle aggregate scoring (v4.2)
        if self.lhandle.layout:
            mod = self.lhandle.layout.compute_aggregate(technique, subscores)
            if mod is not None:
                technique.aggregateScore = mod
        a, b = SVG_Technique(self.lhandle.gradient).build(offset, technique, height,
                                                          width, subtechniques=subtechniques, mode=mode,
                                                          tactic=tactic, colors=colors, tBC=config.tableBorderColor)
        return a, b

    def export(self, showName, showID, lhandle, config, sort=0, scores=[], colors=[], subtechs=[], exclude=[]):
        """
            Export a layer object to an SVG object

            :param showName: Boolean of whether or not to show names
            :param showID:  Boolean of whether or not to show IDs
            :param lhandle: The layer object being exported
            :param config: A SVG Config object
            :param sort: The sort mode
            :param scores: List of tactic scores
            :param colors: List of tactic default colors
            :param subtechs: List of visible subtechniques
            :param exclude: List of excluded techniques
            :return:
        """
        self.codex = self.h.get_matrix(self.mode, filters=lhandle.filters)
        grad = False
        if len(scores):
            grad = lhandle.gradient
        self.lhandle = lhandle
        d, presence, overlay = self._build_headers(lhandle.name, config, lhandle.domain,
                                                   lhandle.versions.attack, lhandle.description, lhandle.filters,
                                                   grad)
        self.codex = self.h._adjust_ordering(self.codex, sort, scores)
        index = 0
        lengths = []
        border = convertToPx(config.border, config.unit)
        glob = G(tx=border)
        for x in self.codex:
            su = len(x.techniques)
            for enum in exclude:
                if enum[0] in [y.id for y in x.techniques]:
                    if self.h.convert(enum[1]) == x.tactic.name or enum[1] is False:
                        su -= 1
            for y in x.subtechniques:
                if y in [z[0] for z in subtechs]:
                    su += len(x.subtechniques[y])
            lengths.append(su)
        tech_width = ((convertToPx(config.width, config.unit) - 2.2 * border) / sum([1 for x in lengths if x > 0])) - \
            border
        header_offset = convertToPx(config.headerHeight, config.unit)
        if presence == 0:
            header_offset = 0
        header_offset += 2.5 * border
        tech_height = (convertToPx(config.height, config.unit) - header_offset -
                       convertToPx(config.border, config.unit)) / (max(lengths) + 1)
        incre = tech_width + 1.1 * border
        for x in self.codex:
            disp = ''
            if showName and showID:
                disp = x.tactic.id + ": " + x.tactic.name
            elif showName:
                disp = x.tactic.name
            elif showID:
                disp = x.tactic.id

            g = G(tx=index, ty=header_offset)

            fs, _ = _optimalFontSize(disp, tech_width, tech_height, maxFontSize=28)
            tx = Text(ctype='TacticName', font_size=fs, text=disp, position='middle')
            gt = G(tx=tech_width/2, ty=tech_height/2)
            gt.append(tx)
            a = self.get_tactic(x, tech_height, tech_width, colors=colors, subtechs=subtechs, exclude=exclude,
                                mode=(showName, showID), scores=scores, config=config)
            b = G(ty=tech_height)
            g.append(gt)
            b.append(a)
            g.append(b)
            if a:
                glob.append(g)
                index += incre
        d.append(glob)
        if overlay:
            d.append(overlay)
        return d

    def _copy_scores(self, listing, scores, tactic_name, exclude):
        """
            INTERNAL: Move scores over from the input object (scores) to the one used to build the svg (listing)

            :param listing: List of objects to apply scores to
            :param scores: List of scores for this tactic
            :param exclude: List of excluded techniques
            :return: None - operates on the raw object itself
        """
        for b in listing:
            if b in exclude:
                b.score = None
                continue
            found = False
            for y in scores:
                if b.id == y[0] and (y[1] == self.h.convert(tactic_name) or not y[1]):
                    b.score = y[2]
                    found = True
                    continue
            if not found:
                b.score = None



import argparse
import copy
import json
import traceback
from uuid import uuid4
from datetime import datetime
from stix2elevator.stix_stepper import step_bundle
from stix2elevator.options import initialize_options, ElevatorOptions

# https://github.com/mitre-attack/attack-stix-data/blob/docs/data-sources/USAGE.md#the-attck-spec
X_MITRE_SPEC_VERSION = "2.1.0"


class STIXToCollection:
    @staticmethod
    def stix_to_collection(bundle, name, version, description=None):
        """Enhance an existing stix bundle with a ATT&CK Collection object
        :param bundle: dictionary representation of a stix bundle
        :param name: name for the generated collection object
        :param version: parameter indicating the ATT&CK version for the generated collection object
        :param description: optional parameter describing the collection
        :returns: updated bundle, now containing a ATT&CK Collection object
        """
        working_bundle = copy.deepcopy(bundle)
        for obj in working_bundle["objects"]:  # check to see if this bundle already contains a collection
            if obj["type"] == 'x-mitre-collection':
                return bundle
        if bundle.get("spec_version", "") == "2.0":
            try:
                print("[NOTE] - version 2.0 spec detected. Forcibly upgrading the bundle to 2.1 to support "
                      "collections.")
                initialize_options(ElevatorOptions(custom_property_prefix='mitre', silent=True))
                working_bundle = step_bundle(working_bundle)
                print("[NOTE] - NOTICE: ATT&CK in STIX 2.1 includes additional fields which were not present on the "
                      "STIX 2.0 data. These fields have not been added automatically and their absence may affect "
                      "compatibility with ingesting software. Please see "
                      "https://github.com/mitre-attack/attack-stix-data/blob/master/USAGE.md for more information.")
            except Exception as e:
                print(f"[ERROR] - Unexpected issue encountered when trying to upgrade from 2.0 to 2.1: {e}. "
                      f"Terminating...")
                print(f"[ERROR] - Full Error trace: {traceback.print_exc(e)}")
                return None
        if bundle.get("spec_version", "") != "2.1" and bundle.get("spec_version", "") != "2.0":
            print(f"[ERROR] - version {working_bundle.get('spec_version', '[NOT FOUND]')} is not one of [2.0, 2.1]. "
                  f"This module only processes stix 2.0 and stix 2.1 bundles.")
        time = datetime.now().strftime("%Y-%m-%dT%H:%M:%S.%fZ")
        if not description:
            description = "This collection was autogenerated by STIXToCollection, as part of mitreattack-python"
        raw_collection = dict(type="x-mitre-collection", id=f"x-mitre-collection--{uuid4()}", spec_version="2.1",
                              name=name, x_mitre_version=version, x_mitre_attack_spec_version=X_MITRE_SPEC_VERSION,
                              description=description, created_by_ref="", created=time, modified=time,
                              object_marking_refs=[], x_mitre_contents=[])
        for obj in working_bundle['objects']:
            if obj['type'] != 'marking-definition':
                try:
                    raw_collection['x_mitre_contents'].append(dict(object_ref=obj["id"],
                                                                   object_modified=obj["modified"]))
                except KeyError as e:
                    print(f"[ERROR] - object {obj} is missing a necessary field: {e}. Exiting this script...")
                    return None
                if "object_marking_refs" in obj.keys():
                    for omr in obj["object_marking_refs"]:
                        if omr not in raw_collection["object_marking_refs"]:
                            raw_collection["object_marking_refs"].append(omr)
                if "created_by_ref" in obj.keys():
                    if obj["created_by_ref"] != raw_collection["created_by_ref"]:
                        if raw_collection["created_by_ref"] != "":
                            print(f"[NOTE] multiple 'created_by_ref' values detected. "
                                  f"{raw_collection['created_by_ref']} (first encountered) will take precedence over "
                                  f"{obj['created_by_ref']}")
                            continue
                        raw_collection["created_by_ref"] = obj["created_by_ref"]

        working_bundle["objects"].insert(0, raw_collection)
        return working_bundle


def main(args):
    with open(args.input, "r", encoding="utf-16") as f:
        bundle = json.load(f)
        with open(args.output, "w", encoding="utf-16") as f2:
            f2.write(json.dumps(STIXToCollection.stix_to_collection(bundle, args.name, args.version, args.description),
                                indent=4))


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Update a STIX 2.0 or 2.1 bundle to include a collection object referencing the contents of the "
                    "bundle."
    )
    parser.add_argument("name",
                        type=str,
                        help="the name for the generated collection object")
    parser.add_argument("version",
                        help="the ATT&CK version for the generated collection object")
    parser.add_argument("--input",
                        type=str,
                        default="bundle.json",
                        help="the input bundle file"
                        )
    parser.add_argument("--output",
                        type=str,
                        default="bundle_out.json",
                        help="the output bundle file"
                        )
    parser.add_argument("--description",
                        type=str,
                        default=None,
                        help="description to use for the generated collection")
    argv = parser.parse_args()
    main(argv)



def handle_object_placement(handle_to_self_field, potential_object, objectType, list=False):
    """
        Helper function to checks to see if we can safely set a given field (handle_to_self_field)
        to a potential_object by checking whether or not it is an instance of objectType.
        :param handle_to_self_field: the field the object is to be assigned to
        :param potential_object: the potential instance of an object
        :param objectType: the type of object being looked for
        :param list: whether or not handle_to_self_field should be a list
        :return: properly referencable object, or False if the object wasn't of the expected type
    """
    if isinstance(potential_object, objectType):
        if list:
            handle_to_self_field.append(potential_object)
        else:
            handle_to_self_field = potential_object
        return handle_to_self_field
    return False



import openpyxl
from openpyxl.styles import Font, Alignment, Border, Side, PatternFill

from mitreattack.navlayers.exporters.matrix_gen import MatrixGen


class BadTemplateException(Exception):
    pass


class ExcelTemplates:

    def __init__(self, source='taxii', resource=None, domain='enterprise'):
        """
            Initialization - Creates a ExcelTemplate object

            :param source: Source to use when compiling the matrix
            :param resource: string path to local cache of stix data (local) or url of workbench to reach out
                                to (remote)
            :param domain: The domain to utilize
        """
        muse = domain
        if muse.startswith('mitre-'):
            muse = domain[6:]
        if muse.endswith('-attack'):
            muse = domain[:-7]
        if muse in ['enterprise', 'mobile']:
            self.mode = muse
            self.h = MatrixGen(source=source, resource=resource)
        else:
            raise BadTemplateException

    def _build_raw(self, showName=True, showID=False, sort=0, scores=[], subtechs=[], exclude=[]):
        """
            INTERNAL - builds a raw, not-yet-marked-up excel document based on the specifications

            :param showName: Whether or not to display names for each entry
            :param showID: Whether or not to display Technique IDs for each entry
            :param sort: The sort mode to use
            :param subtechs: List of all visible subtechniques
            :param exclude: List of of techniques to exclude from the matrix
            :return: a openpyxl workbook object containing the raw matrix
        """
        self.codex = self.h._adjust_ordering(self.codex, sort, scores)
        template, joins = self.h._construct_panop(self.codex, subtechs, exclude)
        self.template = template
        wb = openpyxl.Workbook()

        sheet = wb.active

        header_template_f = Font(name='Calibri', bold=True)
        header_template_a = Alignment(horizontal='center', vertical='bottom')
        header_template_b = Border(bottom=Side(border_style='thin'))
        header_template_c = PatternFill(patternType='solid', start_color='DDDDDD', end_color='DDDDDD')

        for entry in template:
            c = sheet.cell(row=entry[0], column=entry[1])
            write_val = ''
            if showName and showID:
                write_val = self.h._get_ID(self.codex, template[entry]) + ': ' + template[entry]
            elif showName:
                write_val = template[entry]
            elif showID:
                write_val = self.h._get_ID(self.codex, template[entry])
            c.value = write_val
            if entry[0] == 1:
                c.font = header_template_f
                c.alignment = header_template_a
                c.border = header_template_b
                c.fill = header_template_c

        # patch widths
        dims = {}
        sheet_handle = wb.active
        for row in sheet_handle:
            for cell in row:
                if cell.value:
                    dims[cell.column_letter] = max((dims.get(cell.column_letter, 0), len(str(cell.value))))
        for col, value in dims.items():
            sheet_handle.column_dimensions[col].width = value

        merge_border_thickness = 'thin'
        merge_template_l = Border(bottom=Side(border_style=merge_border_thickness),
                                  left=Side(border_style=merge_border_thickness),
                                  top=Side(border_style=merge_border_thickness))
        merge_template_r = Border(bottom=Side(border_style=merge_border_thickness),
                                  right=Side(border_style=merge_border_thickness),
                                  top=Side(border_style=merge_border_thickness))

        for marker in joins:
            sheet_handle.merge_cells(start_row=marker[0], start_column=marker[1], end_row=marker[0] + marker[2] - 1,
                                     end_column=marker[1])
            for block in range(marker[0], marker[0] + marker[2]):
                sheet_handle[block][marker[1]].border = merge_template_r
                sheet_handle[block][marker[1] - 1].border = merge_template_l
            sheet_handle.merge_cells(start_row=1, start_column=marker[1], end_row=1, end_column=marker[1] + 1)
            adjust = sheet_handle.cell(row=marker[0], column=marker[1])
            adjust.alignment = Alignment(vertical='top')

        return wb

    def export(self, showName, showID, filters=None, sort=0, scores=[], subtechs=[], exclude=[]):
        """
            Export a raw customized excel template


            :param showName: Whether or not to display names for each entry
            :param showID: Whether or not to display Technique IDs for each entry
            :param filters: Filter/Platform object to use to filter the matrix
            :param sort: The sort mode to utilize
            :param scores: List of scores to utilize (extracted previously from layer)
            :param subtechs: List of all visible subtechniques
            :param exclude: List of of techniques to exclude from the matrix
            return: a openpyxl workbook object containing the raw matrix
        """
        self.codex = self.h.get_matrix(self.mode, filters=filters)
        return self._build_raw(showName, showID, sort, scores, subtechs, exclude)

    def retrieve_coords(self, techniqueID, tactic=None):
        """
            Locate the openpyxl coordinates of the provided technique for the currently loaded matrix

            :param techniqueID: The ID of the technique to locate
            :param tactic: Optional parameter to isolate the technique to a specific tactic
            :return: A tuple representing the (row, column) of the target element in the workbook
        """
        listing = []
        match = self.h._get_name(self.codex, techniqueID)
        for entry in self.template:
            if self.template[entry] == match:
                if tactic is not None:
                    try:
                        if self.template[(1, entry[1])] != self.h.convert(tactic):
                            continue
                    except KeyError:
                        # account for subtechniques when scanning
                        if self.template[(1, entry[1] - 1)] != self.h.convert(tactic):
                            continue
                listing.append(entry)
        if listing:
            if '.' in techniqueID:
                parent = self.retrieve_coords(techniqueID.split('.')[0], tactic)
                if not parent:
                    return 'HIDDEN'
        return listing



from stix2 import Filter
from itertools import chain
import copy

from mitreattack.navlayers.exporters.matrix_gen import MatrixGen
from mitreattack.navlayers.core.exceptions import BadInput, typeChecker
from mitreattack.navlayers.core.layer import Layer
from mitreattack.navlayers.generators.gen_helpers import remove_revoked_depreciated, get_attack_id, build_data_strings


class UnableToFindStixObject(Exception):
    pass


class StixObjectIsNotValid(Exception):
    pass


class UsageLayerGenerator:
    """Generates a Layer that shows techniques mapped to an input group, software or mitigation"""
    def __init__(self, source, domain='enterprise', resource=None):
        """
        Initialize the Generator
        :param source: Which source to use for data (local, taxii [server], or [remote] ATT&CK Workbench)
        :param domain: Which matrix to use during generation
        :param resource: string path to local STIX data (local) or url of workbench to reach out to (remote)
        """
        self.matrix_handle = MatrixGen(source, resource)
        self.domain = domain
        try:
            self.source_handle = self.matrix_handle.collections[domain]
        except KeyError:
            print(f"[UsageGenerator] - unable to load collection {domain} (current source = {source}).")
            raise BadInput
        self.full_matrix = self.matrix_handle.get_matrix(self.domain)
        self.sources = self.source_handle.query([Filter('type', '=', 'x-mitre-data-source')])
        self.components = self.source_handle.query([Filter('type', '=', 'x-mitre-data-component')])
        self.source_mapping = build_data_strings(self.sources, self.components)

    def get_stix_object(self, match):
        """
        Retrieve the stix object for a given string
        :param match: The string to match on - can be a name, alias, or ATT&CK ID
        :return: the corresponding stix object
        """
        filts = [
            [Filter('name', '=', match)],
            [Filter(match, 'in', 'aliases')],
            [Filter(match, 'in', 'x_mitre_aliases')],
            [Filter('external_references.external_id', '=', match)],
            [Filter('id', '=', match)]  # Support data component type objects from sum generator
        ]
        data = list(chain.from_iterable(self.source_handle.query(f) for f in filts))
        data = remove_revoked_depreciated(data)
        if len(data):
            if len(data) > 1:
                print(f"[Usage Generator] - WARNING! Multiple matches found for {match}: [{data}]. Selecting the first "
                      f"one as default.")
            return data[0]
        raise UnableToFindStixObject

    def get_matrix_data(self, match_pattern):
        """
        Retrieve a list of attack-pattern (technique) objects that map to a group, software or mitigation.
        :param match_pattern: Name, associated group/software (alias), or ATT&CK ID.
                              Techniques mapped to the object matching this pattern are returned.```
        """
        obj = self.get_stix_object(match_pattern)
        if obj['type'] == 'course-of-action':
            verb = 'mitigates'
        elif obj['type'] == 'x-mitre-data-component' or obj['type'] == 'x-mitre-data-source':
            verb = 'detects'
        else:
            verb = 'uses'
        related = self.source_handle.relationships(obj['id'], verb, source_only=True)

        out = self.source_handle.query([
            Filter('type', '=', 'attack-pattern'),
            Filter('id', 'in', [r.target_ref for r in related])
        ])
        return remove_revoked_depreciated(out), obj

    def generate_technique_data(self, raw_matches):
        """
        Generate technique list of dictionary objects (dictionary form of technique listing for a layer)
        :param raw_matches: matching attack-pattern objects
        :return: list of dictionary objects for every technique: score=0 if not in raw_matches, 1 otherwise,
                    description in comments
        """
        shortlist = []
        for match in raw_matches:
            xid = ''
            xphase = ''
            for ref in match.external_references:
                if ref.source_name == 'mitre-attack':
                    xid = ref.external_id
            for phase in match.kill_chain_phases:
                if phase.kill_chain_name == 'mitre-attack':
                    xphase = phase.phase_name
            shortlist.append((xid, xphase, match.description))
        full_matrix_listing = copy.deepcopy(self.full_matrix)
        construct = list()
        for tactic in full_matrix_listing:
            for tech in tactic.techniques:
                construct.append(dict(techniqueID=tech.id, score=0,
                                      tactic=self.matrix_handle.convert(tactic.tactic.name)))
            for tech_key in tactic.subtechniques:
                for subtech in tactic.subtechniques[tech_key]:
                    construct.append(dict(techniqueID=subtech.id, score=0,
                                          tactic=self.matrix_handle.convert(tactic.tactic.name)))
        for entry in shortlist:
            for tac in construct:
                if entry[0] == tac['techniqueID'] and (entry[1] == '' or entry[1] == tac['tactic']):
                    tac['score'] = 1
                    tac['comment'] = entry[2]
        return construct

    def generate_layer(self, match):
        """
        Generate a layer
        :param match: the pattern to match
        :return: layer object with annotated techniques
        """
        typeChecker(type(self).__name__, match, str, "match")
        raw_data, matched_obj = self.get_matrix_data(match)
        if matched_obj['type'] not in ["course-of-action", 'tool', 'malware', 'intrusion-set',
                                       'x-mitre-data-source', 'x-mitre-data-component']:
            print(f"Warning: The input match {match} corresponds with an ATT&CK Technique, which is not supported. "
                  f"Please provide a group, software, or mitigation instead.")
            raise StixObjectIsNotValid
        a_id = get_attack_id(matched_obj)
        processed_listing = self.generate_technique_data(raw_data)
        raw_layer = dict(name=f"{matched_obj['name']} ({matched_obj['id']})", domain=self.domain + '-attack')
        raw_layer['techniques'] = processed_listing
        output_layer = Layer(raw_layer)
        if matched_obj['type'] != 'x-mitre-data-component':
            name = matched_obj['name']
        else:
            name = self.source_mapping[matched_obj['id']]
        output_layer.description = f"{self.domain.capitalize() if len(self.domain) > 3 else self.domain.upper()} " \
                                   f"techniques used by {name}, ATT&CK {matched_obj['type']} {a_id}"
        return output_layer



import copy

import pandas as pd
from stix2 import Filter, MemoryStore
from itertools import chain
from tqdm import tqdm
import datetime
import re
import numpy as np

# Lookup module for Platforms - each matrix has a list of possible platforms, and each platform with multiple
#   subplatforms has a corresponding entry. This allows for a pseudo-recursive lookup of subplatforms, as the presence
#   of a platform at the top level of this lookup indicates the existence of subplatforms.
MATRIX_PLATFORMS_LOOKUP = {"enterprise-attack": ['PRE', 'Windows', 'macOS', 'Linux', 'Cloud', 'Office 365', 'Azure AD',
                                                 'Google Workspace', 'SaaS', 'IaaS', 'Network', 'Containers'],
                           "mobile-attack": ['Android', 'iOS'],
                           "Cloud": ['Office 365', 'Azure AD', 'Google Workspace', 'SaaS', 'IaaS'],
                           "ics-attack": ["Field Controller/RTU/PLC/IED", "Safety Instrumented System/Protection Relay",
                                          "Control Server", "Input/Output Server", "Windows", "Human-Machine Interface",
                                          "Engineering Workstation", "Data Historian"]}

TITLE_EXCLUSIONS = ['and']

def remove_revoked_deprecated(stix_objects):
    """Remove any revoked or deprecated objects from queries made to the data source"""
    # Note we use .get() because the property may not be present in the JSON data. The default is False
    # if the property is not set.
    return list(
        filter(
            lambda x: x.get("x_mitre_deprecated", False) is False and x.get("revoked", False) is False,
            stix_objects
        )
    )


def filter_platforms(stix_objects, platforms):
    """Filter out any objects that don't have a matching platform to one in 'platforms'"""
    if not platforms:
        return stix_objects

    return list(
        filter(
            lambda x: any(platform.lower() in [y.lower() for y in x.get("x_mitre_platforms", [])]
                          for platform in platforms),
            stix_objects
        )
    )


def format_date(date):
    """ Given a date string, return it formatted as %d %B %Y """
    if isinstance(date, str):
        date = datetime.datetime.strptime(date, "%Y-%m-%dT%H:%M:%S.%fZ")
    return "{} {} {}".format(date.strftime("%d"), date.strftime("%B"), date.strftime("%Y"))


def get_citations(objects):
    """given a list of STIX objects, return a pandas dataframe for the citations on the objects"""
    citations = []
    for sdo in objects:
        if "external_references" in sdo:
            for ref in sdo["external_references"]:
                if "external_id" not in ref and "description" in ref and not ref["description"].startswith(
                        "(Citation: "):
                    citation = {
                        "reference": ref["source_name"],
                        "citation": ref["description"]
                    }
                    if "url" in ref:
                        citation["url"] = ref["url"]

                    citations.append(citation)

    return pd.DataFrame(citations).drop_duplicates(subset="reference", ignore_index=True)


def parseBaseStix(sdo):
    """given an SDO, return a dict of field names:values that are common across all ATT&CK STIX types"""
    row = {}
    url = None
    if "external_references" in sdo and sdo["external_references"][0]["source_name"] in ["mitre-attack",
                                                                                         "mitre-mobile-attack",
                                                                                         "mitre-ics-attack"]:
        row["ID"] = sdo["external_references"][0]["external_id"]
        url = sdo["external_references"][0]["url"]
    if "name" in sdo:
        row["name"] = sdo["name"]
    if "description" in sdo:
        row["description"] = sdo["description"]
    if url:
        row["url"] = url
    if "created" in sdo:
        row["created"] = format_date(sdo["created"])
    if "modified" in sdo:
        row["last modified"] = format_date(sdo["modified"])
    if "x_mitre_version" in sdo:
        row["version"] = sdo["x_mitre_version"]
    if "x_mitre_contributors" in sdo:
        row["contributors"] = "; ".join(sorted(sdo["x_mitre_contributors"]))
    return row


def techniquesToDf(src, domain):
    """
    Parse STIX techniques from the given data and return corresponding pandas dataframes.
    :param src: MemoryStore or other stix2 DataSource object holding the domain data
    :param domain: domain of ATT&CK src corresponds to, e.g "enterprise-attack"
    :returns: a lookup of labels (descriptors/names) to dataframes
    """

    techniques = src.query([Filter("type", "=", "attack-pattern")])
    techniques = remove_revoked_deprecated(techniques)
    technique_rows = []

    all_sub_techniques = src.query([
        Filter("type", "=", "relationship"),
        Filter("relationship_type", "=", "subtechnique-of"),
    ])
    all_sub_techniques = MemoryStore(stix_data=all_sub_techniques)

    for technique in tqdm(techniques, desc="parsing techniques"):
        # get parent technique if sub-technique
        subtechnique = "x_mitre_is_subtechnique" in technique and technique["x_mitre_is_subtechnique"]
        if subtechnique:
            subtechnique_of = all_sub_techniques.query([
                Filter("source_ref", "=", technique["id"])
            ])[0]
            parent = src.get(subtechnique_of["target_ref"])

        # base STIX properties
        row = parseBaseStix(technique)

        # sub-technique properties
        tactic_shortnames = list(map(lambda kcp: kcp["phase_name"], technique["kill_chain_phases"]))
        tactics = list(map(lambda t: ' '.join([x.title() if x not in TITLE_EXCLUSIONS else x for x in t.split('-')]),
                           tactic_shortnames))
        row["tactics"] = ", ".join(sorted(tactics))

        if "x_mitre_detection" in technique:
            row["detection"] = technique["x_mitre_detection"]
        if "x_mitre_platforms" in technique:
            row["platforms"] = ", ".join(sorted(technique["x_mitre_platforms"]))

        # domain specific fields -- ICS + Enterprise
        if domain in ["enterprise-attack", "ics-attack"]:
            if "x_mitre_data_sources" in technique:
                row["data sources"] = ", ".join(sorted(technique["x_mitre_data_sources"]))

        # domain specific fields -- enterprise
        if domain == "enterprise-attack":
            row["is sub-technique"] = subtechnique
            if subtechnique:
                row["name"] = f"{parent['name']}: {technique['name']}"
                row["sub-technique of"] = parent["external_references"][0]["external_id"]

            if "x_mitre_system_requirements" in technique:
                row["system requirements"] = ", ".join(sorted(technique["x_mitre_system_requirements"]))
            if "x_mitre_permissions_required" in technique:
                row["permissions required"] = ", ".join(sorted(technique["x_mitre_permissions_required"],
                                                               key=str.lower))
            if "x_mitre_effective_permissions" in technique:
                row["effective permissions"] = ", ".join(sorted(technique["x_mitre_effective_permissions"],
                                                                key=str.lower))

            if "defense-evasion" in tactic_shortnames and "x_mitre_defense_bypassed" in technique:
                row["defenses bypassed"] = ", ".join(sorted(technique["x_mitre_defense_bypassed"]))
            if "execution" in tactic_shortnames and "x_mitre_remote_support" in technique:
                row["supports remote"] = technique["x_mitre_remote_support"]
            if "impact" in tactic_shortnames and "x_mitre_impact_type" in technique:
                row["impact type"] = ", ".join(sorted(technique["x_mitre_impact_type"]))
            capec_refs = list(filter(lambda ref: ref["source_name"] == "capec",
                                     technique["external_references"]))
            if capec_refs:
                row["CAPEC ID"] = ", ".join([x["external_id"] for x in capec_refs])

        # domain specific fields -- mobile
        elif domain == "mobile-attack":
            if "x_mitre_tactic_type" in technique:
                row["tactic type"] = ", ".join(sorted(technique["x_mitre_tactic_type"]))
            mtc_refs = list(filter(lambda ref: ref["source_name"] == "NIST Mobile Threat Catalogue",
                                   technique["external_references"]))
            if mtc_refs:
                row["MTC ID"] = mtc_refs[0]["external_id"]

        technique_rows.append(row)

    citations = get_citations(techniques)
    dataframes = {
        "techniques": pd.DataFrame(technique_rows).sort_values("name"),
    }
    # add relationships
    codex = relationshipsToDf(src, relatedType="technique")
    dataframes.update(codex)
    # add relationship references
    dataframes["techniques"][f"relationship citations"] = _get_relationship_citations(dataframes['techniques'], codex)
    # add/merge citations
    if not citations.empty:
        if "citations" in dataframes:  # append to existing citations from references
            dataframes["citations"] = dataframes["citations"].append(citations)
        else:  # add citations
            dataframes["citations"] = citations

        dataframes["citations"].sort_values("reference")

    return dataframes


def tacticsToDf(src, domain):
    """
    Parse STIX tactics from the given data and return corresponding pandas dataframes.
    :param src: MemoryStore or other stix2 DataSource object holding the domain data
    :param domain: domain of ATT&CK src corresponds to, e.g "enterprise-attack"
    :returns: a lookup of labels (descriptors/names) to dataframes
    """

    tactics = src.query([Filter("type", "=", "x-mitre-tactic")])
    tactics = remove_revoked_deprecated(tactics)

    tactic_rows = []
    for tactic in tqdm(tactics, desc="parsing mitigations"):
        tactic_rows.append(parseBaseStix(tactic))

    citations = get_citations(tactics)
    dataframes = {
        "tactics": pd.DataFrame(tactic_rows).sort_values("name"),
    }
    if not citations.empty: dataframes["citations"] = citations.sort_values("reference")

    return dataframes


def sourcesToDf(src, domain):
    """
    Parse STIX Data Sources and their Data components from the given data and return corresponding pandas dataframes.
    :param src: MemoryStore or other stix2 DataSource object holding the domain data
    :param domain: domain of ATT&CK src corresponds to, e.g "enterprise-attack"
    :returns: a lookup of labels (descriptors/names) to dataframes
    """
    data = list(chain.from_iterable(  # software are the union of the tool and malware types
        src.query(f) for f in [Filter("type", "=", "x-mitre-data-component"),
                               Filter("type", "=", "x-mitre-data-source")]
    ))
    if data:
        refined = remove_revoked_deprecated(data)
        data_object_rows = []
        source_lookup = dict()
        for x in refined:
            if x['type'] == 'x-mitre-data-source':
                source_lookup[x['id']] = x['name']
        for data_object in tqdm(refined, desc="parsing data sources"):
            # add common STIx fields
            row = parseBaseStix(data_object)
            # add software-specific fields
            if "x_mitre_platforms" in data_object:
                row["platforms"] = ", ".join(sorted(data_object["x_mitre_platforms"]))
            if "x_mitre_collection_layers" in data_object:
                row["collection layers"] = ', '.join(sorted(data_object["x_mitre_collection_layers"]))
            if "x_mitre_aliases" in data_object:
                row["aliases"] = ", ".join(sorted(data_object["x_mitre_aliases"][1:]))
            if data_object["type"] == "x-mitre-data-component":
                row["name"] = f"{data_object['name']}: {source_lookup[data_object['x_mitre_data_source_ref']]}"
                row["type"] = "datacomponent"
            else:
                row["type"] = "datasource"
            if "description" in data_object:
                row['description'] = data_object['description']
            data_object_rows.append(row)

        citations = get_citations(refined)
        tempa = pd.DataFrame(data_object_rows).sort_values("name")
        dataframes = {
            "datasources": tempa.reindex(columns=["name", "ID", "description", "collection layers", "platforms", "created",
                                                   "modified", "type", "version", "url", "contributors"]),
        }
        # add relationships
        dataframes.update(relationshipsToDf(src, relatedType="datasource"))
        # add/merge citations
        if not citations.empty:
            if "citations" in dataframes:  # append to existing citations from references
                dataframes["citations"] = dataframes["citations"].append(citations)
            else:  # add citations
                dataframes["citations"] = citations

            dataframes["citations"].sort_values("reference")

        return dataframes
    else:
        print(f'[WARNING] (sourceToDf) - No data components or data sources found for domain {domain}. Skipping...')


def softwareToDf(src, domain):
    """
    Parse STIX software from the given data and return corresponding pandas dataframes.
    :param src: MemoryStore or other stix2 DataSource object holding the domain data
    :param domain: domain of ATT&CK src corresponds to, e.g "enterprise-attack"
    :returns: a lookup of labels (descriptors/names) to dataframes
    """

    software = list(chain.from_iterable(  # software are the union of the tool and malware types
        src.query(f) for f in [Filter("type", "=", "tool"), Filter("type", "=", "malware")]
    ))
    software = remove_revoked_deprecated(software)
    software_rows = []
    for soft in tqdm(software, desc="parsing software"):
        # add common STIx fields
        row = parseBaseStix(soft)
        # add software-specific fields
        if "x_mitre_platforms" in soft:
            row["platforms"] = ", ".join(sorted(soft["x_mitre_platforms"]))
        if "x_mitre_aliases" in soft:
            row["aliases"] = ", ".join(sorted(soft["x_mitre_aliases"][1:]))
        row["type"] = soft["type"]  # malware or tool

        software_rows.append(row)

    citations = get_citations(software)
    dataframes = {
        "software": pd.DataFrame(software_rows).sort_values("name"),
    }
    # add relationships
    codex = relationshipsToDf(src, relatedType="software")
    dataframes.update(codex)
    # add relationship references
    dataframes["software"][f"relationship citations"] = _get_relationship_citations(dataframes["software"], codex)
    # add/merge citations
    if not citations.empty:
        if "citations" in dataframes:  # append to existing citations from references
            dataframes["citations"] = dataframes["citations"].append(citations)
        else:  # add citations
            dataframes["citations"] = citations

        dataframes["citations"].sort_values("reference")

    return dataframes


def groupsToDf(src, domain):
    """
    Parse STIX groups from the given data and return corresponding pandas dataframes.
    :param src: MemoryStore or other stix2 DataSource object holding the domain data
    :param domain: domain of ATT&CK src corresponds to, e.g "enterprise-attack"
    :returns: a lookup of labels (descriptors/names) to dataframes
    """

    groups = src.query([Filter("type", "=", "intrusion-set")])
    groups = remove_revoked_deprecated(groups)
    group_rows = []
    for group in tqdm(groups, desc="parsing groups"):
        row = parseBaseStix(group)
        # add group aliases
        if "aliases" in group:
            associated_groups = []
            associated_group_citations = []
            for alias in sorted(group["aliases"][1:]):
                # find the reference for the alias
                associated_groups.append(alias)
                for ref in group["external_references"]:
                    if ref["source_name"] == alias:
                        associated_group_citations.append(ref["description"])
                        break
                        # aliases.append(alias)
            row["associated groups"] = ", ".join(associated_groups)
            row["associated groups citations"] = ", ".join(associated_group_citations)

        group_rows.append(row)

    citations = get_citations(groups)
    dataframes = {
        "groups": pd.DataFrame(group_rows).sort_values("name"),
    }
    # add relationships
    codex = relationshipsToDf(src, relatedType="group")
    dataframes.update(codex)
    # add relationship references
    dataframes["groups"][f"relationship citations"] = _get_relationship_citations(dataframes["groups"], codex)
    # add/merge citations
    if not citations.empty:
        if "citations" in dataframes:  # append to existing citations from references
            dataframes["citations"] = dataframes["citations"].append(citations)
        else:  # add citations
            dataframes["citations"] = citations

        dataframes["citations"].sort_values("reference")

    return dataframes


def mitigationsToDf(src, domain):
    """
    Parse STIX mitigations from the given data and return corresponding pandas dataframes.
    :param src: MemoryStore or other stix2 DataSource object holding the domain data
    :param domain: domain of ATT&CK src corresponds to, e.g "enterprise-attack"
    :returns: a lookup of labels (descriptors/names) to dataframes
    """

    mitigations = src.query([Filter("type", "=", "course-of-action")])
    mitigations = remove_revoked_deprecated(mitigations)
    mitigation_rows = []
    for mitigation in tqdm(mitigations, desc="parsing mitigations"):
        mitigation_rows.append(parseBaseStix(mitigation))

    citations = get_citations(mitigations)
    dataframes = {
        "mitigations": pd.DataFrame(mitigation_rows).sort_values("name"),
    }
    # add relationships
    codex = relationshipsToDf(src, relatedType="mitigation")
    dataframes.update(codex)
    # add relationship references
    dataframes["mitigations"]["relationship citations"] = _get_relationship_citations(dataframes["mitigations"], codex)
    # add/merge citations
    if not citations.empty:
        if "citations" in dataframes:  # append to existing citations from references
            dataframes["citations"] = dataframes["citations"].append(citations)
        else:  # add citations
            dataframes["citations"] = citations

        dataframes["citations"].sort_values("reference")

    return dataframes


class CellRange:
    """
    helper class for handling ranges of cells in a spreadsheet. Note: not 0-indexed, row and cols start at 1.
    Data is optional argument for data to store in the cellrange in the case of merged ranges
    format is a dict {name, format} for the XlsxWriter style. Formats of the same name will not be defined multiple
        times to the worksheet; only the first definition will be used
    """

    def __init__(self, leftCol, rightCol, topRow, bottomRow, data=None, format=None):
        self.leftCol = leftCol
        self.rightCol = rightCol
        self.topRow = topRow
        self.bottomRow = bottomRow
        self.data = data
        self.format = format

    def to_excel_format(self):
        """return the range in excel format, e.g A4:C7"""
        return f"{self._loc_to_excel(self.topRow, self.leftCol)}:{self._loc_to_excel(self.bottomRow, self.rightCol)}"

    def _loc_to_excel(self, row, col):
        """ Convert given row and column number to an Excel-style cell name. Note: not 0-indexed"""
        letters = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
        result = []
        while col:
            col, rem = divmod(col - 1, 26)
            result[:0] = letters[rem]
        return ''.join(result) + str(row)


def build_technique_and_sub_columns(src, techniques, columns, merge_data_handle, matrix_grid_handle, tactic_name):
    """
    Build technique and subtechnique columns for a given matrix and attach them to the appropriate object listings
    :param src: MemoryStore or other stix2 DataSource object holding the domain data
    :param techniques: List of technique stix objects belong in this tactic column
    :param columns: Existing columns in this matrix (used for placement)
    :param merge_data_handle: Handle to the 'merge' data object for this matrix
    :param matrix_grid_handle: Handle to the 2D grid array being constructed for the matrix (technique and subtechnique
                                columns will be appended here)
    :param tactic_name: The name of the corresponding tactic for this column
    :return: Nothing (meta - modifies the passed in merge_data_handle and matrix_grid_handle objects)
    """
    techniques_column = []
    subtechniques_column = []

    all_sub_techniques = src.query([
        Filter("type", "=", "relationship"),
        Filter("relationship_type", "=", "subtechnique-of"),
    ])
    all_sub_techniques = MemoryStore(stix_data=all_sub_techniques)

    for technique in techniques:
        techniques_column.append(technique["name"])
        # sub-technique relationships
        subtechnique_ofs = all_sub_techniques.query([
            Filter("target_ref", "=", technique["id"])
        ])
        if len(subtechnique_ofs) > 0:  # if there are sub-techniques on the tactic
            technique_top = len(techniques_column) + 1  # top of row range to merge
            # get sub-techniques
            subtechniques = [src.get(rel["source_ref"]) for rel in subtechnique_ofs]
            subtechniques = remove_revoked_deprecated(subtechniques)
            subtechniques = sorted(subtechniques, key=lambda x: x["name"])
            for i in range(len(subtechniques)):  # for each sub-technique
                if i != 0:
                    techniques_column.append("")  # first sub-technique is parallel to the technique in the layout
                subtechniques_column.append(subtechniques[i]["name"])
            technique_bottom = len(techniques_column) + 1  # bottom of row range to merge
            if technique_top != technique_bottom:  # more than 1 sub-technique
                merge_data_handle.append(CellRange(  # merge technique portion of cell group
                    len(columns),
                    len(columns),
                    technique_top,
                    technique_bottom,
                    data=technique["name"],
                    format={  # format of the merged range
                        "name": "supertechnique",
                        "format": {
                            'valign': 'vcenter',
                            'text_wrap': 1,
                            'shrink': 1
                        }
                    }
                ))
        else:  # no sub-techniques; add empty cell parallel to technique
            subtechniques_column.append("")
    # end adding techniques and sub-techniques to column

    matrix_grid_handle.append(techniques_column)  # add technique column to grid

    if len(list(filter(lambda x: x != "", subtechniques_column))) > 0:  # if there are sub-techniques for the tactic
        matrix_grid_handle.append(subtechniques_column)  # add sub-technique sub-column
        columns.append("")  # add empty tactic header for the sub-column
        merge_data_handle.append(  # merge tactic column header with the sub-column header that was just appended
            CellRange(len(columns) - 1,
                      len(columns),
                      1,
                      1,
                      data=tactic_name,
                      format={  # tactic header formatting
                          "name": "tacticHeader",
                          "format": {
                              "bold": 1,
                              "border": 1,
                              "font_size": 14,
                              "align": "center",
                              "shrink": 1
                          }
                      }
                      ))


def build_parsed_DF_matrix(matrix_grid, columns, merge, parsed_dict):
    """
    Build the DF matrix object
    :param matrix_grid: 2D array of the matrix to build
    :param columns: Column headers
    :param merge: Any applicable cell merge ranges and styles
    :param parsed_dict: Dictionary containing name and description for the matrix
    :return: { matrix, name, description, merge, border } where
        matrix is a pandas dataframe of the matrix
        name is the name of the matrix
        description is the description of the matrix
        merge is a list of CellRange objects that need to be merged for formatting of the sub-techniques in the matrix
        columns is the number of columns in the data
    """
    parsed = copy.deepcopy(parsed_dict)
    # reshape array so that pandas consumes it properly
    matrix_grid = np.flip(np.rot90(matrix_grid), 0)
    # create dataframe for array
    df = pd.DataFrame(matrix_grid, columns=columns)

    # Set additional data for the matrix
    parsed["matrix"] = df  # actual dataframe
    parsed["merge"] = merge  # merge ranges and associated styles
    parsed["columns"] = len(columns)  # number of columns with data
    return parsed


def matricesToDf(src, domain):
    """
    Parse STIX matrices from the given data and return parsed matrix structures
    :param src: MemoryStore or other stix2 DataSource object holding the domain data
    :param domain: domain of ATT&CK src corresponds to, e.g "enterprise-attack"
    :returns: [{ matrix, name, description, merge, border }, ... ] where
        matrix is a pandas dataframe of the matrix
        name is the name of the matrix
        description is the description of the matrix
        merge is a list of CellRange objects that need to be merged for formatting of the sub-techniques in the matrix
        columns is the number of columns in the data
    """
    matrices = src.query([Filter("type", "=", "x-mitre-matrix")])
    matrices = remove_revoked_deprecated(matrices)
    matrices_parsed = []
    sub_matrices_parsed = []

    for matrix in tqdm(matrices, desc="parsing matrices"):
        sub_matrices_grid = dict()
        sub_matrices_merges = dict()
        sub_matrices_columns = dict()
        for entry in MATRIX_PLATFORMS_LOOKUP[domain]:
            sub_matrices_grid[entry] = []
            sub_matrices_merges[entry] = []
            sub_matrices_columns[entry] = []

        parsed = {
            "name": matrix['name'] if len(matrices) == 1 else F"{domain.split('-')[0].capitalize()} {matrix['name']}",
            "description": matrix["description"]
        }

        matrix_grid = []  # matrix layout in 2d array
        merge = []  # list of CellRange objects to merge later

        columns = []  # column names
        for tactic_ref in tqdm(matrix["tactic_refs"], desc="processing matrix tactics"):
            tactic = src.get(tactic_ref)
            columns.append(tactic["name"])  # add tactic header

            # parse techniques in tactic
            techniques = list(
                filter(lambda t: not ("x_mitre_is_subtechnique" in t and t["x_mitre_is_subtechnique"]), src.query([
                    Filter("type", "=", "attack-pattern"),
                    Filter("kill_chain_phases.phase_name", "=", tactic["x_mitre_shortname"]),
                ])))
            techniques = remove_revoked_deprecated(techniques)
            techniques = sorted(techniques, key=lambda x: x["name"])
            # add techniques
            build_technique_and_sub_columns(src, techniques, columns, merge, matrix_grid, tactic['name'])

            for platform in MATRIX_PLATFORMS_LOOKUP[domain]:
                # In order to support "groups" of platforms, each platform is checked against the lookup a second time.
                # If an second entry can be found, the results from that query will be used, otherwise, the singular
                # platform will be.
                a_techs = filter_platforms(techniques,
                                           [platform] if platform not in MATRIX_PLATFORMS_LOOKUP
                                           else MATRIX_PLATFORMS_LOOKUP[platform])
                if a_techs:
                    sub_matrices_columns[platform].append(tactic['name'])
                    build_technique_and_sub_columns(src, a_techs, sub_matrices_columns[platform],
                                                    sub_matrices_merges[platform], sub_matrices_grid[platform],
                                                    tactic['name'])

        # square the grid because pandas doesn't like jagged columns
        longest_column = 0
        for column in matrix_grid:
            longest_column = max(len(column), longest_column)
        for column in matrix_grid:
            for i in range((longest_column - len(column))):
                column.append("")

        for submatrix in sub_matrices_grid:
            mg = sub_matrices_grid[submatrix]
            for column in mg:
                longest_column = max(len(column), longest_column)
            for column in mg:
                for i in range((longest_column - len(column))):
                    column.append("")
        # matrix is now squared

        parsed = build_parsed_DF_matrix(matrix_grid, columns, merge, parsed)
        matrices_parsed.append(parsed)

        for submatrix in sub_matrices_grid:
            if sub_matrices_grid[submatrix]:  # make sure we found matches for something
                local = copy.deepcopy(parsed)
                local['name'] = f"{submatrix}" if len(matrices) == 1 else f"{submatrix} {matrix['name']}"
                local['description'] = local['description'].split(":")[0] + f": {submatrix}"
                subparsed = build_parsed_DF_matrix(sub_matrices_grid[submatrix], sub_matrices_columns[submatrix],
                                                   sub_matrices_merges[submatrix], local)
                sub_matrices_parsed.append(subparsed)

    # end adding of matrices
    return matrices_parsed, sub_matrices_parsed


def relationshipsToDf(src, relatedType=None):
    """
    Parse STIX relationships from the given data and return corresponding pandas dataframes.
    :param src: MemoryStore or other stix2 DataSource object holding the domain data
    :param relatedType: optional, singular attack type to only return relationships with, e.g "mitigation"
    :returns: a lookup of labels (descriptors/names) to dataframes
    """

    # Helper lookups
    attackToStixTerm = {
        "technique": ["attack-pattern"],
        "tactic": ["x-mitre-tactic"],
        "software": ["tool", "malware"],
        "group": ["intrusion-set"],
        "mitigation": ["course-of-action"],
        "matrix": ["x-mitre-matrix"],
        "datasource": ["x-mitre-data-component"],
    }
    stixToAttackTerm = {
        "attack-pattern": "technique",
        "x-mitre-tactic": "tactic",
        "tool": "software",
        "malware": "software",
        "intrusion-set": "group",
        "course-of-action": "mitigation",
        "x-mitre-matrix": "matrix",
        "x-mitre-data-component": "datacomponent",
        "x-mitre-data-source": "datasource"
    }

    # get master list of relationships
    relationships = src.query([Filter("type", "=", "relationship")])
    relationships = remove_revoked_deprecated(relationships)
    relationship_rows = []  # build list of rows for dataframe
    # tqdm description depends on the related type and parameters
    iterdesc = "parsing all relationships" if not relatedType else f"parsing relationships for type={relatedType}"
    for relationship in tqdm(relationships, desc=iterdesc):
        source = src.get(relationship["source_ref"])  # source object of the relationship
        target = src.get(relationship["target_ref"])  # target object of the relationship

        # filter if related objects don't exist or are revoked or deprecated
        if not source or source.get("x_mitre_deprecated", False) is True or source.get("revoked", False) is True:
            continue
        if not target or target.get("x_mitre_deprecated", False) is True or target.get("revoked", False) is True:
            continue
        if relationship["relationship_type"] == "revoked":
            continue

        # don't track sub-technique relationships, those are tracked in the techniques df
        if relationship["relationship_type"] == "subtechnique-of":
            continue

        # filter out relationships not with relatedType
        if relatedType:
            related = False
            for stixTerm in attackToStixTerm[relatedType]:  # try all stix types for the ATT&CK type
                if (source["type"] == stixTerm or
                        target["type"] == stixTerm):  # if any stix type is part of the relationship
                    related = True
                    break
            if not related:
                continue  # skip this relationship if the types don't match

        # add mapping data
        row = {}

        def add_side(label, sdo):
            """add data for one side of the mapping"""
            if "external_references" in sdo and sdo["external_references"][0]["source_name"] in ["mitre-attack",
                                                                                                 "mitre-mobile-attack"]:
                row[f"{label} ID"] = sdo["external_references"][0]["external_id"]  # "source ID" or "target ID"
            if "name" in sdo:
                row[f"{label} name"] = sdo["name"]  # "source name" or "target name"
            row[f"{label} type"] = stixToAttackTerm[sdo["type"]]  # "source type" or "target type"

        add_side("source", source)
        row["mapping type"] = relationship["relationship_type"]  # mapping type goes between the source/target data
        add_side("target", target)
        if "description" in relationship:  # add description of relationship to the end of the row
            row["mapping description"] = relationship["description"]

        relationship_rows.append(row)

    citations = get_citations(relationships)
    relationships = pd.DataFrame(relationship_rows).sort_values(
        ["mapping type", "source type", "target type", "source name", "target name"])

    if not relatedType:  # return all relationships and citations
        dataframes = {
            "relationships": relationships,
        }
        if not citations.empty:
            dataframes["citations"] = citations.sort_values("reference")

        return dataframes
    else:  # break into dataframes by mapping type
        dataframes = {}

        # group:software / "associated {other type}"
        relatedGroupSoftware = relationships.query(
            "`mapping type` == 'uses' and (`source type` == 'group' or `source type` == 'software') and "
            "(`target type` == 'group' or `target type` == 'software')")
        if not relatedGroupSoftware.empty:
            dataframes[f"associated {'software' if relatedType == 'group' else 'groups'}"] = relatedGroupSoftware

        # technique:group + technique:software / "procedure examples"
        procedureExamples = relationships.query("`mapping type` == 'uses' and `target type` == 'technique'")
        if not procedureExamples.empty:
            dataframes["procedure examples" if relatedType == "technique" else "techniques used"] = procedureExamples

        # technique:mitigation / "mitigation mappings"
        relatedMitigations = relationships.query("`mapping type` == 'mitigates'")
        if not relatedMitigations.empty:
            dataframes[
                'associated mitigations' if relatedType == 'technique' else 'techniques addressed'] = relatedMitigations

        if not citations.empty:
            # filter citations by ones actually used
            # build master list of used citations
            usedCitations = set()
            for dfname in dataframes:
                df = dataframes[dfname]
                for description in filter(lambda x: x == x, df[
                    "mapping description"].tolist()):  # filter out missing descriptions which for whatever reason
                    # in pandas don't equal themselves
                    [usedCitations.add(x) for x in re.findall(r"\(Citation: (.*?)\)", description)]

            citations = citations[citations.reference.isin(list(usedCitations))]  # filter to only used references

            dataframes["citations"] = citations.sort_values("reference")

        return dataframes


def _get_relationship_citations(object_dataframe, relationship_df):
    """
    Extract citations for each _object_ in the relationship dataframe. This allows us to include
    citations from relationships for each ATT&CK object type.
    :param object_dataframe: Dataframe for relevant ATT&CK object
    :param relationship_df: Dataframe of relationships
    :return: Array of strings, with each string being placed relative to the object listing, and containing all
        relevant citations
    """
    object_listing = [x for x in object_dataframe["ID"]]
    new_citations = []
    for z in [x for x in relationship_df if x != 'citations']:
        subset = []
        for y in object_listing:
            mask = relationship_df[z].values == y
            filtered = relationship_df[z].loc[mask]
            temp = set()
            for description in filter(lambda x: x == x, filtered["mapping description"].tolist()):
                [temp.add(x) for x in re.findall(r"\(Citation: (.*?)\)", description)]
            subset.append(",".join([f"(Citation: {z})" for z in temp]))
        if not new_citations:
            new_citations = subset
        else:
            for i in range(0, len(new_citations)):
                new_citations[i] = ','.join([new_citations[i], subset[i]])
    return new_citations



import argparse
import os
from tqdm import tqdm
import json
import uuid
from dateutil.parser import isoparse
import re
from stix2 import MemoryStore, Filter


class CollectionToIndex:
    @staticmethod
    def generate_index(name, description, root_url, files=None, folders=None, sets=None):
        """generate a collection index from the input data and return the index as a dict
        arguments:
            name (string): the name of the index
            description (string): the description of the index
            root_url (string): the root URL where the collections can be found. Specified collection paths will be
                                appended to this for the collection URL
            files (string[], optional): collection JSON files to include in the index. Cannot be used with folder
                                argument
            folders (string[], optional): folders of collection JSON files to include in the index. Cannot be used with
                                files argument. Will only match collections that end with a version number
            sets (meta[], optional): array of json dictionaries representing stix bundle objects or and array of
                                     MemoryStore objects to include in the index
        """
        if len([x for x in [files, folders, sets] if x]) > 1:
            print("cannot use multiple arguments (files, folders, sets) at the same time, please use only one "
                  "argument at a time")
            return

        if folders:
            version_regex = re.compile("(\w+-)+(\d\.?)+.json")
            files = []
            for folder in folders:
                files += list(map(lambda fname: os.path.join(folder, fname),
                                  filter(lambda fname: version_regex.match(fname), os.listdir(folder))))

        cleaned_bundles = []
        if sets:
            if isinstance(sets[0], MemoryStore):
                uset = [dict(type="bundle", id=f"bundle--{x.id}", spec_version="2.0",
                             objects=x.source.query([Filter("type", "=", "x-mitre-collection")]))
                        for x in sets]
                sets = uset
            for potentially_valid_bundle in sets:
                if potentially_valid_bundle["objects"] is not [[]]:  # Catch case where MemoryStore didn't have a match
                    if any(x["type"] == 'x-mitre-collection' for x in potentially_valid_bundle["objects"]):
                        potentially_valid_bundle["objects"] = filter(lambda x: x["type"] == "x-mitre-collection",
                                                                     potentially_valid_bundle["objects"])
                        cleaned_bundles.append(potentially_valid_bundle)
                    else:
                        print(f"cannot use bundle {potentially_valid_bundle.id} due to lack of collection object")
                else:
                    print(f"cannot use bundle {potentially_valid_bundle.id} due to lack of collection object")

        index_created = None
        index_modified = None
        collections = {}  # STIX ID -> collection object

        if files:
            for collection_bundle_file in tqdm(files, desc="parsing collections"):
                with open(collection_bundle_file, "r", encoding="utf-16") as f:
                    bundle = json.load(f)
                    url = root_url + collection_bundle_file if root_url.endswith("/") \
                        else root_url + "/" + collection_bundle_file
                    CollectionToIndex._extract_collection(bundle, collections, url)

        if cleaned_bundles:
            for input_bundle in tqdm(cleaned_bundles, desc="transferring input bundles"):
                url = "Imported"
                CollectionToIndex._extract_collection(input_bundle, collections, url)

        for collection in collections.values():
            # set collection name and description from most recently modified version
            collection["name"] = collection["versions"][-1]["name"]
            collection["description"] = collection["versions"][-1]["description"]
            # set index created date according to first created collection
            index_created = index_created if index_created and index_created < isoparse(
                collection["created"]) else isoparse(collection["created"])
            # delete name and description from all versions
            for version in collection["versions"]:
                # set index created date according to first created collection
                index_modified = index_modified if index_modified and index_modified > isoparse(
                    version["modified"]) else isoparse(version["modified"])
                # delete name and description from version since they aren't used in the output
                del version["name"]
                del version["description"]

        return {
            "id": str(uuid.uuid4()),
            "name": name,
            "description": description,
            "created": index_created.isoformat(),
            "modified": index_modified.isoformat(),
            "collections": list(collections.values())
        }

    @staticmethod
    def _extract_collection(bundle, collections, url):
        """
        Extract a collection from a bundle, and build it into the passed in collections dictionary
        :param bundle: The bundle to work with
        :param collections: A dictionary to place the extracted collection into
        :param url: The corresponding url for this given collection version
        :return: Nothing (Meta - collections dictionary modified)
        """
        for collection_version in filter(lambda x: x["type"] == "x-mitre-collection", bundle["objects"]):
            # parse collection
            if collection_version["id"] not in collections:
                # create
                collections[collection_version["id"]] = {
                    "id": collection_version["id"],
                    "created": collection_version["created"],  # created is the same for all versions
                    "versions": []
                }
            collection = collections[collection_version["id"]]

            # append this as a version
            collection["versions"].append({
                "version": collection_version["x_mitre_version"],
                "url": url,
                "modified": collection_version["modified"],
                "name": collection_version["name"],  # this will be deleted later in the code
                "description": collection_version["description"],  # this will be deleted later in the code
            })
            # ensure the versions are ordered
            collection["versions"].sort(key=lambda version: isoparse(version["modified"]), reverse=True)


def main(args):
    with open(args.output, "w", encoding="utf-16") as f:
        index = CollectionToIndex.generate_index(name=args.name, description=args.description, root_url=args.root_url,
                                                 files=args.files, folders=args.folders)
        print(f"writing {args.output}")
        json.dump(index, f, indent=4, enforce_ascii=False)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Create a collection index from a set of collections"
    )
    parser.add_argument(
        "name",
        type=str,
        default=None,
        help="name of the collection index. If omitted a placeholder will be used"
    )
    parser.add_argument(
        "description",
        type=str,
        default=None,
        help="description of the collection index. If omitted a placeholder will be used"
    )
    parser.add_argument(
        "root_url",
        type=str,
        help="the root URL where the collections can be found. Specified collection paths will be appended to this for "
             "the collection URL"
    )
    parser.add_argument(
        "--output",
        type=str,
        default="index.json",
        help="filename for the output collection index file"
    )
    input_options = parser.add_mutually_exclusive_group(required=True)  # require at least one input type
    input_options.add_argument(
        '--files',
        type=str,
        nargs="+",
        default=None,
        metavar=("collection1", "collection2"),
        help="list of collections to include in the index"
    )
    input_options.add_argument(
        '--folders',
        type=str,
        nargs="+",
        default=None,
        help="folder of JSON files to treat as collections"
    )

    argv = parser.parse_args()
    main(argv)



import colour
import math

from mitreattack.navlayers.core.exceptions import typeChecker, typeCheckerArray


class Gradient:
    def __init__(self, colors, minValue, maxValue):
        """
            Initialization - Creates a gradient object

            :param colors: The array of color codes for this gradient
            :param minValue: The minValue for this gradient
            :param maxValue: The maxValue for this gradient
        """
        self.__minValue = None
        self.__maxValue = None
        self.__alpha = []
        self.colors = colors
        self.minValue = minValue
        self.maxValue = maxValue

    @property
    def colors(self):
        return self.__colors

    @colors.setter
    def colors(self, colors):
        typeCheckerArray(type(self).__name__, colors, str, "colors")
        self.__colors = []
        self.__alpha = []
        for entry in colors:
            try:
                colour.Color(entry)
            except AttributeError:
                self.__alpha.append(entry[-2:])
                if entry.startswith('#'):
                    entry = entry[:7]
                else:
                    entry = entry[:6]
            self.__colors.append(entry)
        self._compute_curve()

    @property
    def minValue(self):
        return self.__minValue

    @minValue.setter
    def minValue(self, minValue):
        typeChecker(type(self).__name__, minValue, int, "minValue")
        self.__minValue = minValue
        self._compute_curve()

    @property
    def maxValue(self):
        return self.__maxValue

    @maxValue.setter
    def maxValue(self, maxValue):
        typeChecker(type(self).__name__, maxValue, int, "maxValue")
        self.__maxValue = maxValue
        self._compute_curve()

    def _compute_curve(self):
        """
            Computes the gradient color curve
        """
        if self.maxValue is not None and self.minValue is not None and self.colors is not None:
            chunksize = int(math.floor((self.maxValue - self.minValue)/(len(self.colors) - 1)))
            fchunksize = int(math.ceil((self.maxValue - self.minValue)/(len(self.colors) - 1)))
            self.curve = []
            index = 1
            while index < len(self.colors):
                s_c = colour.Color(self.colors[index-1])
                e_c = colour.Color(self.colors[index])
                if index == len(self.colors):
                    curve_2 = list(s_c.range_to(e_c, fchunksize))
                else:
                    curve_2 = list(s_c.range_to(e_c, chunksize))
                index += 1
                self.curve.extend(curve_2)
            self.curve.append(colour.Color(self.colors[-1]))

    def compute_color(self, score):
        """
            Computes a specific color based on the score value provided
            :returns: A hexadecimal color representation of the score on
                the gradient
        """
        if score <= self.minValue:
            return self.curve[0].hex_l
        if score >= self.maxValue:
            return self.curve[-1].hex_l

        target = self.curve[int(score - self.minValue)]
        return target.hex_l

    def get_dict(self):
        """
            Converts the currently loaded gradient file into a dict
            :returns: A dict representation of the current gradient object
        """
        if len(self.__alpha) == len(self.__colors):
            color_out = [f"{self.__colors[i]}{self.__alpha[i]}" for i in range(0, len(self.__alpha))]
        else:
            color_out = self.__colors
        return dict(colors=color_out, minValue=self.__minValue,
                    maxValue=self.maxValue)



import drawSvg
import colorsys
import numpy as np
import os
from PIL import ImageFont

from mitreattack.navlayers.core.gradient import Gradient


def convertToPx(quantity, unit):
    """
        INTERNAL: Convert values to pixels

        :param quantity: value
        :param unit: unit for that value
        :return: quantity in pixels
    """
    if unit == "in":
        return quantity * 96
    if unit == "cm":
        return quantity * 37.79375
    if unit == "px":
        return quantity
    if unit == "em":
        return quantity * 16
    if unit == "pt":
        return quantity * 1.33
    return -1


def _getstringwidth(string, font, size):
    """
        INTERNAL: Calculate the width of a string (in pixels)

        :param string: string to evaluate
        :param font: font to use
        :param size: font size
        :return: pixel length of string
    """
    font = ImageFont.truetype('{}/fonts/{}.ttf'.format(os.path.sep.join(__file__.split(os.path.sep)[:-1]), font),
                              int(size))
    length, _ = font.getsize(string)
    return length


def _getstringheight(string, font, size):
    """
        INTERNAL: Calculate the width of a string (in pixels)

        :param string: string to evaluate
        :param font: font to use
        :param size: font size
        :return: pixel height of string
    """
    font = ImageFont.truetype('{}/fonts/{}.ttf'.format(os.path.sep.join(__file__.split(os.path.sep)[:-1]), font),
                              int(size))
    _, height = font.getsize(string)
    return height


def _findSpace(words, width, height, maxFontSize):
    """
        INTERNAL: Find space locations for a string to keep it within width x height

        :param words: string to evaluate
        :param width: width of the box
        :param height: height of the box
        :param maxFontSize: maximum font size
        :return:
    """
    padding = 4
    breakDistance = min(height, (maxFontSize + 3) * len(words))

    breakTextHeight = breakDistance / len(words)
    fitTextHeight = min(breakTextHeight, height) * 0.8

    longestWordLength = -9999
    fitTextWidth = 9999
    for w in range(0, len(words)):
        word = words[w]
        longestWordLength = max(longestWordLength, len(word))
    try:
        fitTextWidth = ((width - (2 * padding)) / longestWordLength * 1.45)
    except ZeroDivisionError:
        pass
    size = min(maxFontSize, fitTextHeight, fitTextWidth)
    return size


def _find_breaks(num_spaces, num_breaks=3):
    """
        INTERNAL: Generate break mapping

        :param num_spaces: number of spaces in string
        :param num_breaks: number of breaks to insert
        :return: list of possible break mappings
    """
    breaks = set()

    def recurse(breakset_inherit, depth, break_count):
        """recursive combinatorics; breakset is binary array of break locations; depth is the depth of recursion,
            break_count is how many breaks should be added"""
        for i in range(len(breakset_inherit)):  # for each possible break
            # insert a break here
            breakset = np.copy(breakset_inherit)
            breakset[i] = 1
            breaks.add("".join(str(x) for x in breakset))
            # try inserting another depth of break
            if depth < break_count - 1:
                recurse(breakset, depth + 1, break_count)

    initial_breaks = [0] * num_spaces
    breaks.add("".join(str(x) for x in initial_breaks))
    recurse(initial_breaks, 0, num_breaks)

    return breaks


def _optimalFontSize(st, width, height, maxFontSize=12):
    """
        INTERNAL: Calculate the optimal fontsize and word layout for a box of width x height

        :param st: string to fit
        :param width: box width
        :param height: box height
        :param maxFontSize: maximum allowable font size
        :return: size in pixels for font, array of strings split by where new lines should go
    """
    words = st.split(" ")
    bestSize = -9999
    bestWordArrangement = []

    num_spaces = len(words)
    num_breaks = 1
    if num_spaces < 20:
        num_breaks = 2
    elif num_spaces < 50:
        num_breaks = 3

    if _findSpace([st], width, height, maxFontSize) == maxFontSize:
        return maxFontSize, [st]
    breaks = _find_breaks(num_spaces, num_breaks)
    for binaryString in breaks:
        wordSet = []

        for k in range(0, len(binaryString)):
            if binaryString[k] == "0":
                if len(wordSet) == 0:
                    wordSet.append(words[k])
                else:
                    wordSet[len(wordSet) - 1] = wordSet[len(wordSet) - 1] + " " + words[k]
            else:
                wordSet.append(words[k])

        size = _findSpace(wordSet, width, height, maxFontSize)
        if size > bestSize:
            bestSize = size
            bestWordArrangement = wordSet

    return bestSize, bestWordArrangement


class Cell(drawSvg.DrawingParentElement):
    TAG_NAME = 'rect'

    def __init__(self, height, width, fill, tBC, ctype=None):
        # tBC = tableBorderColor, ctype='class' field on resulting svg object, fill=[R,G,B]
        super().__init__(height=height, width=width, style='fill: rgb({}, {}, {})'.format(fill[0], fill[1], fill[2]),
                         stroke=tBC)
        if ctype:
            self.args['class'] = ctype


class HeaderRect(drawSvg.DrawingParentElement):
    TAG_NAME = 'rect'

    def __init__(self, width, height, ctype, x=None, y=None, outline=True):
        # ctype='class' field on resulting svg object, x=x coord, y=y coord
        super().__init__(width=width, height=height, fill='white', rx='5')
        self.args['class'] = ctype
        if x:
            self.args['x'] = x
        if y:
            self.args['y'] = y
        if outline:
            self.args['stroke'] = 'black'


class G(drawSvg.DrawingParentElement):
    TAG_NAME = 'g'

    def __init__(self, tx=None, ty=None, style=None, ctype=None):
        # tx=translate x, ty=translate y, ctype='class' field on resulting svg object
        super().__init__()
        if tx is None:
            tx = 0
        if ty is None:
            ty = 0
        self.args['transform'] = 'translate(' + str(tx) + ',' + str(ty) + ')'
        if style:
            self.args['style'] = style
        if ctype:
            self.args['class'] = ctype


class Line(drawSvg.DrawingParentElement):
    TAG_NAME = 'line'

    def __init__(self, x1, x2, y1, y2, stroke):
        # x1=start x, x2=stop x, y1=start y, y2=stop y, stroke='stroke' field on resulting svg object
        super().__init__(x1=x1, x2=x2, y1=y1, y2=y2, stroke=stroke)


class Text(drawSvg.Text):
    def __init__(self, text, font_size, ctype, position=None, tx=None, ty=None, x=None, y=None, fill=None):
        # ctype='class' object on resulting svg, position='text-anchor' field, tx/ty=translate x/y, x/y=x/y coord
        if x is None:
            x = 0
        if y is None:
            y = 0
        super().__init__(text=text, fontSize=font_size, x=x, y=-y)
        self.args['class'] = ctype
        if tx is None:
            tx = 0
        if ty is None:
            ty = 0
        if tx != 0 or ty != 0:
            self.args['transform'] = 'translate(' + str(tx) + ',' + str(ty) + ')'
        if position:
            self.args['style'] = 'text-anchor: {}'.format(position)
        if fill:
            self.args['fill'] = fill


class Swatch(drawSvg.DrawingParentElement):
    TAG_NAME = 'rect'

    def __init__(self, height, width, fill):
        # fill= [R,G,B]
        super().__init__(height=height, width=width, style='fill: rgb({}, {}, {})'.format(fill[0], fill[1], fill[2]))


class SVG_HeaderBlock:
    @staticmethod
    def build(height, width, label, config, variant='text', t1text=None, t2text=None, colors=[]):
        """
            Build a single SVG Header Block object

            :param height: Height of the block
            :param width: Width of the block
            :param label: Label for the block
            :param config: SVG configuration object
            :param variant: text or graphic - the type of header block to build
            :param t1text: upper text
            :param t2text: lower text
            :param colors: array of tuple (color, score value) for the graphic variant
            :return:
        """
        g = G(ty=5)
        rect = HeaderRect(width, height, 'header-box')
        g.append(rect)
        rect2 = HeaderRect(_getstringwidth(label, config.font, 12), _getstringheight(label, config.font, 12),
                           'label-cover', x=7, y=-5, outline=False)
        g.append(rect2)
        text = Text(label, 12, 'header-box-label', x=8, y=3)
        g.append(text)
        internal = G(tx=5, ctype='header-box-content')
        g.append(internal)
        if variant == 'text':
            upper = G(tx=0, ty=2.1)
            internal.append(upper)
            if t1text is not None:
                bu = t2text is not None and t2text != ""
                theight = (height-8)
                if bu:
                    theight = theight / 2
                fs, patch_text = _optimalFontSize(t1text, width, theight, maxFontSize=28)
                lines = len(patch_text)
                y = theight/2 + 2.1
                if lines > 1:
                    y = y - (theight / 5 * (lines - 1) - (fs * 6/16))
                if float(fs) < (convertToPx(config.border, config.unit) + 2.1):
                    y = theight/2 + 2.1 - (theight / 5)
                t1 = Text("\n".join(patch_text), fs, '', x=4, y=y)
                upper.append(t1)
                if bu:
                    upper.append(Line(0, width - 10, theight, theight, stroke='#dddddd'))
                    upper_fs = fs
                    lower_offset = (theight + 2.1)
                    lower = G(tx=0, ty=lower_offset)
                    fs, patch_text = _optimalFontSize(t2text, width, (height - (height/3 + upper_fs)), maxFontSize=28)
                    y = theight / 2 + 5.1
                    lines = len(patch_text)
                    adju = "\n".join(patch_text)
                    if lines > 1:
                        y = y - ((theight / 5) * (lines - 1))
                    if float(fs) > lower_offset:
                        y = y + 2*(float(fs) - lower_offset)
                    t2 = Text(adju, fs, '', x=4, y=y)
                    lower.append(t2)
                    internal.append(lower)
        else:
            if len(colors):
                usable = width - 10
                block_width = usable/len(colors)
                sub1 = G(ty=5)
                sub2 = G(ty=5)
                internal.append(sub1)
                sub1.append(sub2)
                cells = G(ctype="legendCells")
                sub2.append(cells)
                offset = 0
                for entry in colors:
                    cell = G(ctype="cell", tx=offset)
                    conv = entry[0]
                    if conv.startswith('#'):
                        conv = conv[1:]
                    block = Swatch(15, block_width, tuple(int(conv[i:i+2], 16) for i in (0, 2, 4)))
                    offset += block_width
                    cell.append(block)
                    cells.append(cell)
                    tblob = str(entry[1])
                    off = (block_width-(5 * (1 + len(tblob))))/2
                    if off < 0:
                        off = 0
                    fs, _ = _optimalFontSize("0", width/len(colors), height)
                    label = Text(tblob, fs, ctype='label', ty=25, tx=off)
                    cell.append(label)
        return g


class SVG_Technique:
    def __init__(self, gradient):
        self.grade = gradient
        if self.grade is None:
            self.grade = Gradient(colors=["#ff6666", "#ffe766", "#8ec843"], minValue=1, maxValue=100)

    def build(self, offset, technique, height, width, tBC, subtechniques=[], mode=(True, False), tactic=None,
              colors=[]):
        """
            Build a SVG Technique block

            :param offset: Current offset to build the block at (so it fits in the column)
            :param technique: The technique to build a block for
            :param height: The height of the technique block
            :param width: The width of the technique block
            :param tBC: The hex code of the technique block's border
            :param subtechniques: List of any visible subtechniques for this technique
            :param mode: Display mode (Show Name, Show ID)
            :param tactic: The corresponding tactic
            :param colors: List of all default color values if no score can be found
            :return: The newly created SVG technique block
        """
        g = G(ty=offset)
        c = self._com_color(technique, tactic, colors)
        t = dict(name=self._disp(technique.name, technique.id, mode), id=technique.id,
                 color=tuple(int(c[i:i+2], 16) for i in (0, 2, 4)))
        tech, text = self._block(t, height, width, tBC=tBC)
        g.append(tech)
        g.append(text)
        new_offset = height
        for entry in subtechniques:
            gp = G(tx=width/5, ty=new_offset)
            g.append(gp)
            c = self._com_color(entry, tactic, colors)
            st = dict(name=self._disp(entry.name, entry.id, mode), id=entry.id,
                      color=tuple(int(c[i:i + 2], 16) for i in (0, 2, 4)))
            subtech, subtext = self._block(st, height, width - width/5, tBC=tBC)
            gp.append(subtech)
            gp.append(subtext)
            new_offset = new_offset + height
        if len(subtechniques):
            g.append(drawSvg.Lines(width/16, -height,
                                   width/8, -height * 2,
                                   width/8, -height * (len(subtechniques) + 1),
                                   width/5, -height * (len(subtechniques) + 1),
                                   width/5, -height,
                                   close=True,
                                   fill=tBC,
                                   stroke=tBC))
        return g, offset + new_offset

    @staticmethod
    def _block(technique, height, width, tBC):
        """
            INTERNAL: Build a technique block element

            :param technique: Technique data dictionary
            :param height: Block height
            :param width: Block width
            :param tBC: Block border color
            :return: Block object, fit text object
        """
        tech = Cell(height, width, technique['color'], ctype=technique['id'], tBC=tBC)

        fs, patch_text = _optimalFontSize(technique['name'], width, height)
        adjusted = "\n".join(patch_text)

        lines = adjusted.count('\n')

        y = height / 2
        if lines > 0:
            y = (height - (lines * fs)) / 2 + height/10  # padding
        else:
            y = y + fs / 4

        hls = colorsys.rgb_to_hls(technique['color'][0], technique['color'][1], technique['color'][2])
        fill = None
        if hls[1] < 127.5:
            fill = 'white'

        text = Text(adjusted.encode('utf-8').decode('ascii', 'backslashreplace'), fs, '', x=4, y=y, fill=fill)
        return tech, text

    def _com_color(self, technique, tactic, colors=[]):
        """
            INTERNAL: Retrieve hex color for a block

            :param technique: Technique object
            :param tactic: What tactic the technique falls under
            :param colors: Default technique color data
            :return: Hex color code
        """
        c = 'FFFFFF'
        tscore = None
        if technique.score is not None:
            tscore = technique.score
        if hasattr(technique, 'aggregateScore'):
            tscore = technique.aggregateScore
        if tscore is not None:
            c = self.grade.compute_color(tscore)[1:]
        else:
            for x in colors:
                if x[0] == technique.id and (x[1] == tactic or not x[1]):
                    c = x[2][1:]
        return c

    @staticmethod
    def _disp(name, id, mode):
        """
            INTERNAL: Generate technique display form

            :param name: The name of the technique
            :param id: The ID of the technique
            :param mode: Which mode to use
            :return: Target display string for the technique
        """
        p1 = name
        p2 = id
        if not mode[0]:
            p1 = ''
        if not mode[1]:
            p2 = ''
        out = ': '.join([p2, p1])
        if out.startswith(': '):
            return p1
        return out



from mitreattack.navlayers.core.exceptions import typeChecker


class Link:
    def __init__(self, label, url):
        """
            Initialization - Creates a link object

            :param label: the label for this link entry
            :param url: the corresponding url for this link entry
        """
        self.label = label
        self.url = url

    @property
    def label(self):
        return self.__label

    @label.setter
    def label(self, label):
        typeChecker(type(self).__name__, label, str, "label")
        self.__label = label

    @property
    def url(self):
        return self.__url

    @url.setter
    def url(self, url):
        typeChecker(type(self).__name__, url, str, "url")
        self.__url = url

    def get_dict(self):
        """
            Converts the currently loaded data into a dict
            :returns: A dict representation of the local metadata object
        """
        return dict(label=self.__label, url=self.__url)


class LinkDiv:
    def __init__(self, active):
        """
            Initialization - Creates a Link object divider
        """
        self.__name = "DIVIDER"
        self.__value = active

    @property
    def name(self):
        return self.__name

    @property
    def state(self):
        return self.__value

    @state.setter
    def state(self, state):
        typeChecker(type(self).__name__, state, bool, "state")
        self.__value = state

    def get_dict(self):
        """
            Converts the currently loaded data into a dict
            :returns: A dict representation of the local metadata object
        """
        return dict(name=self.__name, value=self.__value)



from mitreattack.navlayers.core.exceptions import typeChecker


class LegendItem:
    def __init__(self, label, color):
        """
            Initialization - Creates a legendItem object

            :param label: The label described by this object
            :param color: The color associated with the label
        """
        self.label = label
        self.color = color

    @property
    def color(self):
        return self.__color

    @color.setter
    def color(self, color):
        typeChecker(type(self).__name__, color, str, "color")
        self.__color = color

    @property
    def label(self):
        return self.__label

    @label.setter
    def label(self, label):
        typeChecker(type(self).__name__, label, str, "label")
        self.__label = label

    def get_dict(self):
        """
            Converts the currently loaded data into a dict
            :returns: A dict representation of the local legendItem object
        """
        return dict(label=self.__label, color=self.__color)



from mitreattack.navlayers.core.exceptions import typeChecker


class Metadata:
    def __init__(self, name, value):
        """
            Initialization - Creates a metadata object

            :param name: the name for this metadata entry
            :param value: the corresponding value for this metadata entry
        """
        self.name = name
        self.value = value

    @property
    def name(self):
        return self.__name

    @name.setter
    def name(self, name):
        typeChecker(type(self).__name__, name, str, "name")
        self.__name = name

    @property
    def value(self):
        return self.__value

    @value.setter
    def value(self, value):
        if isinstance(value, bool):
            value = str(value)
        typeChecker(type(self).__name__, value, str, "value")
        self.__value = value

    def get_dict(self):
        """
            Converts the currently loaded data into a dict
            :returns: A dict representation of the local metadata object
        """
        return dict(name=self.__name, value=self.__value)


class MetaDiv:
    def __init__(self, active):
        """
            Initialization - Creates a metadata object divider
        """
        self.__name = "DIVIDER"
        self.__value = active

    @property
    def name(self):
        return self.__name

    @property
    def state(self):
        return self.__value

    @state.setter
    def state(self, state):
        typeChecker(type(self).__name__, state, bool, "state")
        self.__value = state

    def get_dict(self):
        """
            Converts the currently loaded data into a dict
            :returns: A dict representation of the local metadata object
        """
        return dict(name=self.__name, value=self.__value)



import argparse
from stix2 import MemoryStore
import os
import requests
import pandas as pd

import mitreattack.attackToExcel.stixToDf as stixToDf


INVALID_CHARACTERS = ["\\", "/", "*", "[", "]", ":", "?"]
SUB_CHARACTERS = ["\\", "/"]


def get_stix_data(domain, version=None, remote=None):
    """
    download the ATT&CK STIX data for the given domain and version from MITRE/CTI (or just domain if a remote workbench
    is specified).
    :param domain: the domain of ATT&CK to fetch data from, e.g "enterprise-attack"
    :param version: the version of attack to fetch data from, e.g "v8.1". If omitted, returns the latest version
                    (not used for invocations that use remote)
    :param remote: optional url to a ATT&CK workbench instance. If specified, data will be retrieved from the target
                    Workbench instead of MITRE/CTI
    :returns: a MemoryStore containing the domain data
    """
    if remote:  # Using Workbench Instance
        if ':' not in remote[6:]:
            remote += ":3000"
        if not remote.startswith('http'):
            remote = 'http://' + remote
        url = f"{remote}/api/stix-bundles?domain={domain}&includeRevoked=true&includeDeprecated=true"
        stix_json = requests.get(url).json()
        return MemoryStore(stix_json)
    else:  # Using MITRE/CTI
        if version:
            url = f"https://raw.githubusercontent.com/mitre/cti/ATT%26CK-{version}/{domain}/{domain}.json"
        else:
            url = f"https://raw.githubusercontent.com/mitre/cti/master/{domain}/{domain}.json"

        stix_json = requests.get(url).json()
        return MemoryStore(stix_data=stix_json["objects"])


def build_dataframes(src, domain):
    """
    build pandas dataframes for each attack type, and return a dictionary lookup for each type to the relevant dataframe
    :param src: MemoryStore or other stix2 DataSource object
    :param domain: domain of ATT&CK src corresponds to, e.g "enterprise-attack"
    :returns: a dict lookup of each ATT&CK type to dataframes for the given type to be ingested by write_excel
    """
    df = {
            "techniques": stixToDf.techniquesToDf(src, domain),
            "tactics": stixToDf.tacticsToDf(src, domain),
            "software": stixToDf.softwareToDf(src, domain),
            "groups": stixToDf.groupsToDf(src, domain),
            "mitigations": stixToDf.mitigationsToDf(src, domain),
            "matrices": stixToDf.matricesToDf(src, domain),
            "relationships": stixToDf.relationshipsToDf(src)
        }
    # get each ATT&CK type
    if domain == 'enterprise-attack':
        df["datasources"] = stixToDf.sourcesToDf(src, domain)
    return df


def write_excel(dataframes, domain, version=None, outputDir="."):
    """
    given a set of dataframes from build_dataframes, write the ATT&CK dataset to output directory
    :param dataframes: pandas dataframes as built by build_dataframes
    :param domain: domain of ATT&CK the dataframes correspond to, e.g "enterprise-attack"
    :param version: optional, the version of ATT&CK the dataframes correspond to, e.g "v8.1".
                    If omitted, the output files will not be labelled with the version number
    :param outputDir: optional, the directory to write the excel files to. If omitted writes to a
                      subfolder of the current directory depending on specified domain and version
    :returns: a list of filepaths corresponding to the files written by the function
    """

    print("writing formatted files... ", end="", flush=True)
    # master list of files that have been written
    written_files = []
    # set up output directory
    if version:
        domainVersionString = f"{domain}-{version}"
    else:
        domainVersionString = domain
    outputDirectory = os.path.join(outputDir, domainVersionString)
    if not os.path.exists(outputDirectory):
        os.makedirs(outputDirectory)
    # master dataset file
    master_fp = os.path.join(outputDirectory, f"{domainVersionString}.xlsx")
    master_writer = pd.ExcelWriter(master_fp, engine='xlsxwriter')
    citations = pd.DataFrame()  # master list of citations
    # write individual dataframes and add to master writer
    for objType in dataframes:
        if objType != "matrices":
            # write the dataframes for the object type into named sheets
            fp = os.path.join(outputDirectory, f"{domainVersionString}-{objType}.xlsx")
            obj_writer = pd.ExcelWriter(fp)
            for dfname in dataframes[objType]:
                dataframes[objType][dfname].to_excel(obj_writer, sheet_name=dfname, index=False)
            obj_writer.save()
            written_files.append(fp)

            # add citations to master citations list
            if "citations" in dataframes[objType]:
                citations = citations.append(dataframes[objType]["citations"])

            # add main df to master dataset
            dataframes[objType][objType].to_excel(master_writer, sheet_name=objType, index=False)
        else:  # handle matrix special formatting
            fp = os.path.join(outputDirectory, f"{domainVersionString}-{objType}.xlsx")
            matrix_writer = pd.ExcelWriter(fp, engine='xlsxwriter')
            combined = dataframes[objType][0] + dataframes[objType][1]  # Combine both matrix types
            for matrix in combined:  # some domains have multiple matrices
                # name them accordingly if there are multiple
                sheetname = "matrix" if len(combined) == 1 else matrix["name"] + " matrix"
                for character in INVALID_CHARACTERS:
                    sheetname = sheetname.replace(character, " or " if character in SUB_CHARACTERS else " ")

                if len(sheetname) > 31:
                    sheetname = sheetname[0:28] + "..."
                listing = []
                if matrix in dataframes[objType][0]:  # avoid printing subtype matrices to the master file
                    matrix["matrix"].to_excel(master_writer, sheet_name=sheetname,
                                              index=False)  # write unformatted matrix data to master file
                    listing.append(master_writer)

                matrix["matrix"].to_excel(matrix_writer, sheet_name=sheetname,
                                          index=False)  # write unformatted matrix to matrix file
                listing.append(matrix_writer)

                # for each writer, format the matrix for readability
                for writer in listing:
                    # define column border styles
                    borderleft = writer.book.add_format({"left": 1, "shrink": 1})
                    borderright = writer.book.add_format({"right": 1, "shrink": 1})
                    formats = {}  # formats only need to be defined once: pointers stored here for subsequent uses
                    sheet = writer.sheets[sheetname]

                    sheet.set_column(0, matrix["columns"],
                                     width=20)  # set all columns to 20 width, and add text shrinking to fit

                    # merge supertechniques and tactic headers if sub-techniques are present on a tactic
                    for mergeRange in matrix["merge"]:
                        if mergeRange.format:  # sometimes merge ranges have formats to add to the merged range
                            if mergeRange.format["name"] not in formats:  # add format to book if not defined
                                formats[mergeRange.format["name"]] = writer.book.add_format(mergeRange.format["format"])
                            theformat = formats[mergeRange.format["name"]]  # get saved format if already added
                            # tactic header merge has additional behavior
                            if mergeRange.format["name"] == "tacticHeader":
                                # also set border for entire column for grouping
                                sheet.set_column(
                                    mergeRange.leftCol - 1,
                                    mergeRange.leftCol - 1,
                                    width=20,  # set column widths to make matrix more readable
                                    cell_format=borderleft  # left border around tactic
                                )
                                sheet.set_column(
                                    mergeRange.rightCol - 1,
                                    mergeRange.rightCol - 1,
                                    width=20,  # set column widths to make matrix more readable
                                    cell_format=borderright  # right border around tactic
                                )
                        else:
                            theformat = None  # no format
                        sheet.merge_range(mergeRange.to_excel_format(), mergeRange.data, theformat)  # apply the merge

            matrix_writer.save()  # save the matrix data
            written_files.append(fp)
            # end of matrix sheet writing

    # remove duplicate citations and add sheet to master file
    citations.drop_duplicates(subset="reference", ignore_index=True).sort_values("reference")\
        .to_excel(master_writer, sheet_name="citations", index=False)
    # write the master file
    master_writer.save()
    written_files.append(master_fp)
    print("done")
    print("files created:")
    for thefile in written_files:
        print("\t", thefile)
    return written_files


def export(domain="enterprise-attack", version=None, outputDir=".", remote=None):
    """
    Download ATT&CK data from MITRE/CTI and convert it to excel spreadsheets
    :param domain: the domain of ATT&CK to download, e.g "enterprise-attack"
    :param version: optional, the version of ATT&CK to download, e.g "v8.1". If omitted will build the current version
                    of ATT&CK
    :param outputDir: optional, the directory to write the excel files to. If omitted writes to a
                      subfolder of the current directory depending on specified domain and version
    :param remote: optional, the URL of a remote ATT&CK Workbench instance to connect to for stix data
    """
    # build dataframes
    dataframes = build_dataframes(get_stix_data(domain, version, remote), domain)
    write_excel(dataframes, domain, version, outputDir)


def main():
    parser = argparse.ArgumentParser(
        description="Download ATT&CK data from MITRE/CTI and convert it to excel spreadsheets"
    )
    parser.add_argument("-domain",
                        type=str,
                        choices=["enterprise-attack", "mobile-attack", "ics-attack"],
                        default="enterprise-attack",
                        help="which domain of ATT&CK to convert"
                        )
    parser.add_argument("-version",
                        type=str,
                        help="which version of ATT&CK to convert. If omitted, builds the latest version"
                        )
    parser.add_argument("-output",
                        type=str,
                        default=".",
                        help="output directory. If omitted writes to a subfolder of the current directory depending on "
                             "the domain and version"
                        )
    parser.add_argument("-remote",
                        type=str,
                        default=None,
                        help="remote url of an ATT&CK workbench server. If omitted, stix data will be acquired from the"
                             " official ATT&CK Taxii server (cti-taxii.mitre.org)"
                        )
    args = parser.parse_args()

    export(args.domain, args.version, args.output, args.remote)


if __name__ == '__main__':
    main()



import json
import traceback

from mitreattack.navlayers.core.exceptions import UninitializedLayer, BadType, BadInput, handler
from mitreattack.navlayers.core.layerobj import _LayerObj


class Layer:
    def __init__(self, init_data={}, name=None, domain=None, strict=True):
        """
             Initialization - create a new Layer object
             :param init_data: Optionally provide base Layer json or string
                data on initialization
         """
        self.__layer = None
        self.strict = strict
        if isinstance(name, str) and isinstance(domain, str):
            self.__data = dict(name=name, domain=domain)
            self._build()
        elif isinstance(init_data, str):
            self.from_str(init_data)
        else:
            self.from_dict(init_data)

    @property
    def layer(self):
        if self.__layer is not None:
            return self.__layer
        return "No Layer Loaded Yet!"

    @layer.setter
    def layer(self, layer):
        if isinstance(layer, _LayerObj):
            self.__layer = layer

    def from_str(self, init_str):
        """
            Loads a raw layer string into the object
            :param init_str: the string representing the layer data to
                be loaded
        """
        self.__data = json.loads(init_str)
        self._build()

    def from_dict(self, init_dict):
        """
            Loads a raw layer string into the object
            :param init_dict: the dictionary representing the layer data to
                be loaded
        """
        self.__data = init_dict
        if self.__data != {}:
            self._build()

    def from_file(self, filename):
        """
             loads input from a layer file specified by filename
             :param filename: the target filename to load from
        """
        fallback = False
        with open(filename, 'r', encoding='utf-16') as fio:
            try:
                raw = fio.read()
            except UnicodeError or UnicodeDecodeError:
                fallback = True
        if fallback:
            with open(filename, 'r')as fio:
                raw = fio.read()
        self.__data = json.loads(raw)
        self._build()

    def to_file(self, filename):
        """
            saves the current state of the layer to a layer file specified by
                filename
            :param filename: the target filename to save as
        """
        if self.__layer is not None:
            with open(filename, 'w', encoding='utf-16') as fio:
                json.dump(self.__layer.get_dict(), fio, ensure_ascii=False)
        else:
            raise UninitializedLayer

    def _build(self):
        """
            Loads the data stored in self.data into a LayerObj (self.layer)
        """
        try:
            self.__layer = _LayerObj(self.__data['name'],  self.__data['domain'])
        except BadType or BadInput as e:
            handler(type(self).__name__, 'Layer is malformed: {}. '
                                         'Unable to load.'.format(e))
            self.__layer = None
            return
        except KeyError as e:
            handler(type(self).__name__, 'Layer is missing parameters: {}. '
                                         'Unable to load.'.format(e))
            self.__layer = None
            return

        for key in self.__data:
            if key not in ['name', 'domain']:
                try:
                    self.__layer._linker(key, self.__data[key])
                except Exception as e:
                    if self.strict:
                        handler(type(self).__name__, "{} encountered [{}]. "
                                                     "Unable to load."
                                .format(str(e.__class__.__name__), e))
                        handler(type(self).__name__, "Full Traceback - {}"
                                .format(traceback.format_exc()))
                        self.__layer = None
                        return

    def to_dict(self):
        """
            Converts the currently loaded layer file into a dict
            :returns: A dict representation of the current layer object
        """
        if self.__layer is not None:
            return self.__layer.get_dict()

    def to_str(self):
        """
            Converts the currently loaded layer file into a string
                representation of a dictionary
            :returns: A string representation of the current layer object
        """
        if self.__layer is not None:
            return json.dumps(self.to_dict())



MITRE_ATTACK_DOMAIN_STRINGS = ['mitre-attack', 'mitre-mobile-attack', 'mitre-ics-attack']


def remove_revoked_depreciated(listing):
    """
    Remove revoked and depreciated elements from the listing
    :param listing: input element list
    :return: input element list - revoked elements
    """
    return list(
        filter(
            lambda x: x.get("x_mitre_deprecated", False) is False and x.get("revoked", False) is False,
            listing
        )
    )


def construct_relationship_mapping(mapping_obj, rel):
    """
    Helper to safely add an relationship to a listing in the mapping_obj
    :param mapping_obj: A dictionary containing mappings of attack object types (course-of-action,
                        tool, malware, or intrusion-set) to relationships
    :param rel: The relationship to add
    :return: Updated mapping_obj
    """
    if rel['target_ref'] not in mapping_obj:
        mapping_obj[rel['target_ref']] = []
    mapping_obj[rel['target_ref']].append(rel)


def get_attack_id(obj):
    """
    Helper function to get the ATT&CK ID from an object
    :param obj: The object to extract from
    :return: The ATT&CK ID in string form
    """
    if not obj['type'].startswith('x-mitre'):
        for entry in obj['external_references']:
            if entry['source_name'] in MITRE_ATTACK_DOMAIN_STRINGS:
                return entry['external_id']
    else:
        return obj['id']
    return '-1'


def build_data_strings(data_sources, data_components):
    """
    Build source->component strings for layer generation
    :param data_sources: List of Data Sources (dicts)
    :param data_components: List of Data Components (dicts)
    :return: dict mapping of Data Component IDs to generated source->component strings
    """
    out = dict()
    for component in data_components:
        ref = component['x_mitre_data_source_ref']
        try:
            source = [x for x in data_sources if x['id'] == ref][0]
            name = f"{source['name']}: {component['name']}"
            out[component['id']] = name
        except IndexError:
            print(f'[generator] - (WARNING): no matching data source{ref} found for data component {component["id"]}. '
                  f'Skipping...')
            pass
    return out



from mitreattack.navlayers.core.exceptions import typeCheckerArray, categoryChecker, UNSETVALUE


class Filter:
    def __init__(self, domain="enterprise-attack"):
        """
            Initialization - Creates a filter object, with an optional
                domain input

            :param domain: The domain used for this layer (mitre-enterprise
                or mitre-mobile)
        """
        self.domain = domain
        self.__platforms = UNSETVALUE

    @property
    def platforms(self):
        if self.__platforms != UNSETVALUE:
            return self.__platforms

    @platforms.setter
    def platforms(self, platforms):
        typeCheckerArray(type(self).__name__, platforms, str, "platforms")
        self.__platforms = []
        for entry in platforms:
            self.__platforms.append(entry)

    def get_dict(self):
        """
            Converts the currently loaded data into a dict
            :returns: A dict representation of the local filter object
        """
        temp = dict()
        listing = vars(self)
        for entry in listing:
            if entry == 'domain':
                continue
            if listing[entry] != UNSETVALUE:
                subname = entry.split('__')[-1]
                if subname != 'stages':
                    temp[subname] = listing[entry]
        if len(temp) > 0:
            return temp


class Filterv3(Filter):
    def __init__(self, domain="mitre-enterprise"):
        self.__stages = UNSETVALUE
        super().__init__(domain)

    @property
    def stages(self):
        if self.__stages != UNSETVALUE:
            return self.__stages

    @stages.setter
    def stages(self, stage):
        typeCheckerArray(type(self).__name__, stage, str, "stage")
        categoryChecker(type(self).__name__, stage[0], ["act", "prepare"],
                        "stages")
        self.__stages = stage



UNSETVALUE = '(x)'


class BadInput(Exception):
    pass


class BadType(Exception):
    pass


class UninitializedLayer(Exception):
    pass


class UnknownLayerProperty(Exception):
    pass


class UnknownTechniqueProperty(Exception):
    pass


class MissingParameters(Exception):
    pass


def handler(caller, msg):
    """
        Prints a debug/warning/error message
        :param caller: the entity that called this function
        :param msg: the message to log
    """
    print('[{}] - {}'.format(caller, msg))


def typeChecker(caller, testee, desired_type, field):
    """
        Verifies that the tested object is of the correct type
        :param caller: the entity that called this function (used for error
            messages)
        :param testee: the element to test
        :param desired_type: the type the element should be
        :param field: what the element is to be used as (used for error
            messages)
        :raises BadType: error denoting the testee element is not of the
            correct type
    """
    if not isinstance(testee, desired_type):
        handler(caller, '{} [{}] is not a {}'.format(testee, field,
                                                     str(desired_type)))
        raise BadType


def typeCheckerArray(caller, testee, desired_type, field):
    """
        Verifies that the tested object is an array of the correct type
        :param caller: the entity that called this function (used for error
            messages)
        :param testee: the element to test
        :param desired_type: the type the element should be
        :param field: what the element is to be used as (used for error
            messages)
        :raises BadType: error denoting the testee element is not of the
            correct type
    """
    if not isinstance(testee, list):
        handler(caller, '{} [{}] is not a {}'.format(testee, field, "Array"))
        raise BadType
    if not isinstance(testee[0], desired_type):
        handler(caller, '{} [{}] is not a {}'.format(testee, field, "Array of " + desired_type))
        raise BadType


def categoryChecker(caller, testee, valid, field):
    """
        Verifies that the tested object is one of a set of valid values
        :param caller: the entity that called this function (used for error
            messages)
        :param testee: the element to test
        :param valid: a list of valid values for the testee
        :param field: what the element is to be used as (used for error
            messages)
        :raises BadInput: error denoting the testee element is not one of
            the valid options
    """
    if testee not in valid:
        handler(caller, '{} not a valid value for {}'.format(testee, field))
        raise BadInput


def loadChecker(caller, testee, required, field):
    """
        Verifies that the tested object contains all required fields
        :param caller: the entity that called this function (used for error
            messages)
        :param testee: the element to test
        :param required: a list of required values for the testee
        :param field: what the element is to be used as (used for error
            messages)
        :raises BadInput: error denoting the testee element is not one of
            the valid options
    """
    for entry in required:
        if entry not in testee:
            handler(caller, '{} is not present in {} [{}]'.format(entry, field, testee))
            raise MissingParameters



from stix2 import TAXIICollectionSource, Filter, MemoryStore
from stix2.datastore.memory import _add
from taxii2client.v20 import Server, Collection
import requests
import json


class DomainNotLoadedError(Exception):
    pass


class MatrixEntry:
    def __init__(self, id=None, name=None, platforms=[]):
        if id is not None:
            self.id = id
        if name is not None:
            self.name = name
        self.__platforms = []
        self.platforms = platforms
        self.score = None

    @property
    def id(self):
        if self.__id is not None:
            return self.__id

    @id.setter
    def id(self, new_id):
        self.__id = new_id

    @property
    def name(self):
        if self.__name is not None:
            return self.__name

    @name.setter
    def name(self, name):
        self.__name = name

    @property
    def platforms(self):
        if self.__platforms is not None:
            return self.__platforms

    @platforms.setter
    def platforms(self, platforms):
        if isinstance(platforms, list):
            self.__platforms.extend(platforms)
        else:
            self.__platforms.append(platforms)

    @property
    def score(self):
        if self.__score is not None:
            return self.__score

    @score.setter
    def score(self, score):
        self.__score = score


class Tactic:
    def __init__(self, tactic=None, techniques=None, subtechniques=None):
        if tactic is not None:
            self.tactic = tactic
        if techniques is not None:
            self.techniques = techniques
        if subtechniques is not None:
            self.subtechniques = subtechniques

    @property
    def tactic(self):
        if self.__tactic is not None:
            return self.__tactic

    @tactic.setter
    def tactic(self, tactic):
        self.__tactic = tactic

    @property
    def techniques(self):
        if self.__techniques is not None:
            return self.__techniques

    @techniques.setter
    def techniques(self, techniques):
        self.__techniques = techniques

    @property
    def subtechniques(self):
        if self.__subtechniques is not None:
            return self.__subtechniques

    @subtechniques.setter
    def subtechniques(self, subtechniques):
        self.__subtechniques = subtechniques


class MatrixGen:
    def __init__(self, source='taxii', resource=None):
        """
            Initialization - Creates a matrix generator object

            :param source: Source to utilize (taxii, remote, or local)
            :param resource: string path to local cache of stix data (local) or url of an ATT&CK Workbench (remote)
        """
        self.convert_data = {}
        self.collections = dict()
        if source.lower() not in ['taxii', 'local', 'remote']:
            print('[MatrixGen] - Unable to generate matrix, source {} is not one of "taxii", "remote" or '
                  '"local"'.format(source))
            raise ValueError

        if source.lower() == 'taxii':
            self.server = Server('https://cti-taxii.mitre.org/taxii')
            self.api_root = self.server.api_roots[0]
            for collection in self.api_root.collections:
                if collection.title != "PRE-ATT&CK":
                    tc = Collection('https://cti-taxii.mitre.org/stix/collections/' + collection.id)
                    self.collections[collection.title.split(' ')[0].lower()] = TAXIICollectionSource(tc)
        elif source.lower() == 'local':
            if resource is not None:
                hd = MemoryStore()
                hd.load_from_file(resource)
                if 'mobile' in resource.lower():
                    self.collections['mobile'] = hd
                else:
                    self.collections['enterprise'] = hd
            else:
                print('[MatrixGen] - "local" source specified, but path to local source not provided')
                raise ValueError
        elif source.lower() == 'remote':
            if resource is not None:
                if ':' not in resource[6:]:
                    print('[MatrixGen] - "remote" source missing port; assuming ":3000"')
                    resource += ":3000"
                if not resource.startswith('http'):
                    resource = 'http://' + resource
                for dataset in ['enterprise', 'mobile']:
                    hd = MemoryStore()
                    response = requests.get(f"{resource}/api/stix-bundles?domain={dataset}-"
                                            f"attack&includeRevoked=true&includeDeprecated=true")
                    response.raise_for_status()  # ensure we notice bad responses
                    _add(hd, json.loads(response.text), True, None)
                    self.collections[dataset] = hd
            else:
                print(f'[MatrixGen] - WARNING: "remote" selected without providing a "resource" url. The use of '
                      f'"remote" requires the inclusion of a "resource" url to an ATT&CK Workbench instance. No matrix '
                      f'will be generated...')
        self.matrix = {}
        self._build_matrix()

    @staticmethod
    def _remove_revoked_deprecated(content):
        """Remove any revoked or deprecated objects from queries made to the data source"""
        return list(
            filter(
                lambda x: x.get("x_mitre_deprecated", False) is False and x.get("revoked", False) is False,
                content
            )
        )

    def _search(self, domain, query):
        interum = self.collections[domain].query(query)
        return self._remove_revoked_deprecated(interum)

    def _get_tactic_listing(self, domain='enterprise'):
        """
            INTERNAL - retrieves tactics for the associated domain

            :param domain: The domain to draw from
        """
        tactics = {}
        t_filt = []
        matrix = self._search(domain, [Filter('type', '=', 'x-mitre-matrix')])
        for i in range(len(matrix)):
            tactics[matrix[i]['name']] = []
            for tactic_id in matrix[i]['tactic_refs']:
                tactics[matrix[i]['name']].append(self._search(domain, ([Filter('id', '=', tactic_id)]))[0])
        for entry in tactics[matrix[0]['name']]:
            self.convert_data[entry['x_mitre_shortname']] = entry['name']
            self.convert_data[entry['name']] = entry['x_mitre_shortname']
            t_filt.append(MatrixEntry(id=entry['external_references'][0]['external_id'], name=entry['name']))
        return t_filt

    def _get_technique_listing(self, tactic, domain='enterprise'):
        """
            INTERNAL - retrieves techniques for a given tactic and domain

            :param tactic: The tactic to grab techniques from
            :param domain: The domain to draw from
        """
        techniques = []
        subtechs = {}
        techs = self._search(domain, [Filter('type', '=', 'attack-pattern'),
                                      Filter('kill_chain_phases.phase_name', '=', tactic)])
        for entry in techs:
            if entry['kill_chain_phases'][0]['kill_chain_name'] == 'mitre-attack' or \
                            entry['kill_chain_phases'][0]['kill_chain_name'] == 'mitre-mobile-attack':
                tid = [t['external_id'] for t in entry['external_references'] if 'attack' in t['source_name']]
                platform_tags = []
                if 'x_mitre_platforms' in entry:
                    platform_tags = entry['x_mitre_platforms']
                if '.' not in tid[0]:
                    techniques.append(MatrixEntry(id=tid[0], name=entry['name'], platforms=platform_tags))
                else:
                    parent = tid[0].split('.')[0]
                    if parent not in subtechs:
                        subtechs[parent] = []
                    subtechs[parent].append(MatrixEntry(id=tid[0], name=entry['name'], platforms=platform_tags))
        return techniques, subtechs

    def _adjust_ordering(self, codex, mode, scores=[]):
        """
            INTERNAL - Adjust ordering of matrix based on sort mode

            :param codex: The pre-existing matrix data
            :param mode: The sort mode to use
            :param scores: Any relevant scores to use in modes 2, 3
        """
        if mode == 0:
            return codex
        if mode == 1:
            for colm in codex:
                colm.technique.reverse()
                for sub in colm.subtechniques:
                    colm.subtechniques[sub].reverse()
            return codex
        for colm in codex:
            for st in colm.subtechniques:
                for sub in colm.subtechniques[st]:
                    sub.score = 0
                    for entry in scores:
                        if entry[0] == sub.id and (entry[1] is False or entry[1] == self.convert(colm.tactic.name)):
                            sub.score = entry[2]
                            break
            for tech in colm.techniques:
                tech.score = 0
                for entry in scores:
                    if entry[0] == tech.id and (entry[1] is False or entry[1] == self.convert(colm.tactic.name)):
                        tech.score = entry[2]
                        break
        if mode == 2:
            for colm in codex:
                for tsub in colm.subtechniques:
                    colm.subtechniques[tsub].sort(key=lambda x: x.score)
                colm.techniques.sort(key=lambda x: x.score)
        if mode == 3:
            for colm in codex:
                for tsub in colm.subtechniques:
                    colm.subtechniques[tsub].sort(key=lambda x: x.score, reverse=True)
                colm.techniques.sort(key=lambda x: x.score, reverse=True)
        return codex

    def _construct_panop(self, codex, subtechs, excludes):
        """
            INTERNAL - Creates a list of lists template for the matrix layout

            :param codex: A list of lists matrix (output of .get_matrix())
            :param subtechs: A list of subtechniques that will be visible
            :param excludes: A list of techniques that will be excluded
        """
        st = [x[0] for x in subtechs]
        s_tacs = [x[1] for x in subtechs]
        et = [x[0] for x in excludes]
        e_tacs = [x[1] for x in excludes]

        matrix_obj = {}
        column = 0
        cycle = False
        to_add = []
        stechs = []
        joins = []
        for col in codex:
            # each column of the matrix
            column += 1
            if cycle:
                for entry in to_add:
                    sr = entry[0]
                    joins.append([entry[0], column-1, len(stechs[entry[1]])])
                    for element in stechs[entry[1]]:
                        matrix_obj[(sr, column)] = element.name
                        sr += 1
                cycle = False
                column += 1
            row = 2
            matrix_obj[(1, column)] = col.tactic.name
            c_name = col.tactic.name
            stechs = col.subtechniques
            to_add = []
            for element in col.techniques:
                elname = element.name
                tid = element.id
                skip = False
                for entry in range(0, len(et)):
                    if et[entry] == tid and (e_tacs[entry] is False or self.convert(e_tacs[entry]) == c_name):
                        skip = True
                        break
                if not skip:
                    matrix_obj[(row, column)] = elname
                    sat = False
                    for entry in range(0, len(st)):
                        if st[entry] == tid and (s_tacs[entry] is False or self.convert(s_tacs[entry]) == c_name):
                            # this tech has enabled subtechs
                            to_add.append((row, tid))
                            row += len(stechs[tid])
                            cycle = True
                            sat = True
                            break
                    if not sat:
                        row += 1
        return matrix_obj, joins

    def _get_ID(self, codex, name):
        """
            INTERNAL - Do lookups to retrieve the ID of a technique given it's name

            :param codex: The list of lists matrix object (output of get_matrix)
            :param name: The name of the technique to retrieve the ID of
            :return: The ID of the technique referenced by name
        """
        for col in codex:
            if col.tactic.name == name:
                return col.tactic.id
            for entry in col.subtechniques:
                for subtech in col.subtechniques[entry]:
                    if subtech.name == name:
                        return subtech.id
            for entry in col.techniques:
                if entry.name == name:
                    return entry.id
        return ''

    def _get_name(self, codex, id):
        """
            INTERNAL - Do lookups to retrieve the name of a technique given it's ID

            :param codex: The list of lists matrix object (output of get_matrix)
            :param id: The ID of the technique to retrieve the name of
            :return: The name of the technique referenced by id
        """
        for col in codex:
            if col.tactic.id == id:
                return col.tactic.name
            for entry in col.subtechniques:
                for subtech in col.subtechniques[entry]:
                    if subtech.id == id:
                        return subtech.name
            for entry in col.techniques:
                if entry.id == id:
                    return entry.name
        return ''

    def convert(self, input_str):
        """
            Convert tactic names to and from short names

            :param input_str: A tactic normal or short name
            :return: The tactic's short or normal name
        """
        if self.convert_data == {}:
            return None
        if input_str in self.convert_data:
            return self.convert_data[input_str]

    def _build_matrix(self, domain='enterprise'):
        """
            Build a ATT&CK matrix object, as a list of lists containing technique dictionaries

            :param domain: The domain to build a matrix for
        """
        if domain not in self.collections:
            raise DomainNotLoadedError
        self.matrix[domain] = []
        tacs = self._get_tactic_listing(domain)
        for tac in tacs:
            techs, subtechs = self._get_technique_listing(tac.name.lower().replace(' ', '-'), domain)
            stemp = {}
            # sort subtechniques via id, append to column
            for par in subtechs:
                subtechs[par].sort(key=lambda x: x.name)
                stemp[par] = subtechs[par]
            # sort techniques alphabetically, append to column
            techs.sort(key=lambda x: x.name)
            colm = Tactic(tactic=tac, techniques=techs, subtechniques=stemp)
            self.matrix[domain].append(colm)

    def get_matrix(self, domain='enterprise', filters=None):
        """
            Retrieve an ATT&CK Domain object

            :param domain: The domain to build a matrix for
            :param filters: Any platform filters to apply to the matrix
        """
        if domain not in self.matrix:
            self._build_matrix(domain)
        return self._filter_matrix_platforms(self.matrix[domain], filters)

    @staticmethod
    def _filter_matrix_platforms(matrix, filters):
        """
            INTERNAL - Filter a matrix according to its platforms
            :param matrix: the matrix to refine
            :param filters: a list of platforms to filter
            :return: filtered matrix
        """
        if filters:
            filter_platforms = [x.lower() for x in filters.platforms]
            new_matrix = []
            for tac in matrix:
                ntech_list = []
                nsubtech_list = {}
                for tech in tac.techniques:
                    if any(x.lower() in filter_platforms for x in tech.platforms):
                        ntech_list.append(tech)
                for tech_subs in tac.subtechniques:
                    temp_list = []
                    for subtech in tac.subtechniques[tech_subs]:
                        if any(x.lower() in filter_platforms for x in subtech.platforms):
                            temp_list.append(subtech)
                    if temp_list:
                        nsubtech_list[tech_subs] = temp_list
                if ntech_list:
                    ntac = Tactic(tactic=tac.tactic, techniques=ntech_list, subtechniques=nsubtech_list)
                    new_matrix.append(ntac)
            if new_matrix:
                return new_matrix
            else:
                print("[WARNING] - Unable to produced filtered matrix... nothing would be left under these platform"
                      " restrictions.")
                return matrix
        else:
            return matrix



from enum import Enum

from mitreattack.navlayers.core.exceptions import typeChecker, categoryChecker, UNSETVALUE, handler


class Aggregates(Enum):
    average = 1
    min = 2
    max = 3
    sum = 4


class Layout:
    def __init__(self):
        """
            Initialization - Creates a layout object
        """
        self.__layout = UNSETVALUE
        self.__showID = UNSETVALUE
        self.__showName = UNSETVALUE
        self.__showAggregateScores = UNSETVALUE
        self.__countUnscored = UNSETVALUE
        self.__aggregateFunction = Aggregates.average

    def compute_aggregate(self, technique, subtechniques):
        """
            Compute the aggregate score for a technique and any subtechniques
            :param technique: The chosen technique object
            :param subtechniques: Any relevant subtechnique objects
            :return: Integer representing aggregate score (or None if not computable)
        """
        scores = []
        total = 0
        if self.showAggregateScores:
            if technique.score is not None:
                scores = [technique.score]
                total += 1
            elif self.countUnscored:
                total += 1
            for x in subtechniques:
                if x.score is not None:
                    scores.append(x.score)
                    total += 1
                elif self.countUnscored:
                    total += 1
            if not scores:
                return None
            if all(x == 0 for x in scores):
                return 0
            modified = self._aggFunction(scores, total)
            return modified
        else:
            return None

    def _aggFunction(self, data_block, total_techs):
        """
            Apply the selected aggregate function to a data_block
            :param data_block: List containing scores to compute over (generated by compute_aggregate)
            :param total_techs: Count of number of techniques utilized
            :return: Calculated score
        """
        if self.showAggregateScores:
            data = data_block
            if self.__aggregateFunction == Aggregates.average:
                if total_techs < len(data_block):
                    handler(type(self).__name__, "Total elements > Expected Length? Defaulting to Total Elements")
                    total_techs = len(data_block)
                return sum(data)/total_techs
            elif self.__aggregateFunction == Aggregates.min:
                return min(data)
            elif self.__aggregateFunction == Aggregates.max:
                return max(data)
            elif self.__aggregateFunction == Aggregates.sum:
                return sum(data)

    @property
    def layout(self):
        if self.__layout != UNSETVALUE:
            return self.__layout

    @layout.setter
    def layout(self, layout):
        typeChecker(type(self).__name__, layout, str, "layout")
        categoryChecker(type(self).__name__, layout, ["side", "flat", "mini"],
                        "layout")
        self.__layout = layout

    @property
    def showID(self):
        if self.__showID != UNSETVALUE:
            return self.__showID

    @showID.setter
    def showID(self, showID):
        typeChecker(type(self).__name__, showID, bool, "showID")
        self.__showID = showID

    @property
    def showName(self):
        if self.__showName != UNSETVALUE:
            return self.__showName

    @showName.setter
    def showName(self, showName):
        typeChecker(type(self).__name__, showName, bool, "showName")
        self.__showName = showName

    @property
    def showAggregateScores(self):
        if self.__showAggregateScores != UNSETVALUE:
            return self.__showAggregateScores

    @showAggregateScores.setter
    def showAggregateScores(self, showAggregateScores):
        typeChecker(type(self).__name__, showAggregateScores, bool,
                    "showAggregateScores")
        self.__showAggregateScores = showAggregateScores

    @property
    def countUnscored(self):
        if self.__countUnscored != UNSETVALUE:
            return self.__countUnscored

    @countUnscored.setter
    def countUnscored(self, countUnscored):
        typeChecker(type(self).__name__, countUnscored, bool, "countUnscored")
        self.__countUnscored = countUnscored

    @property
    def aggregateFunction(self):
        if self.__aggregateFunction == Aggregates.average:
            return "average"
        elif self.__aggregateFunction == Aggregates.min:
            return "min"
        elif self.__aggregateFunction == Aggregates.max:
            return "max"
        elif self.__aggregateFunction == Aggregates.sum:
            return "sum"

    @aggregateFunction.setter
    def aggregateFunction(self, aggregateFunction):
        categoryChecker(type(self).__name__, aggregateFunction.lower(),
                        ["average", "min", "max", "sum"], "aggregateFunction")
        self.__aggregateFunction = Aggregates[aggregateFunction.lower()]

    def get_dict(self):
        """
            Converts the currently loaded data into a dict
            :returns: A dict representation of the local layout object
        """
        listing = vars(self)
        temp = dict()
        for entry in listing:
            if entry.split(type(self).__name__ + '__')[-1] == 'aggregateFunction':
                temp['aggregateFunction']\
                    = self.__getAggstring()
            elif listing[entry] != UNSETVALUE:
                temp[entry.split(type(self).__name__ + '__')[-1]]\
                    = listing[entry]
        if len(temp) > 0:
            return temp

    def __getAggstring(self):
        if self.__aggregateFunction == Aggregates.average:
            return 'average'
        if self.__aggregateFunction == Aggregates.min:
            return 'min'
        if self.__aggregateFunction == Aggregates.max:
            return 'max'
        if self.__aggregateFunction == Aggregates.sum:
            return 'sum'
